/* xrmap - vector based global map generating program 
 *
 * Copyright (C) 2001 Jean-Pierre Demailly <demailly@ujf-grenoble.fr>
 * ftp://ftp.ac-grenoble.fr/ge/geosciences
 *
 * derived from "rmap"
 * Copyright (C) 2000 Reza Naima <reza@reza.net>
 * http://www.reza.net/rmap/
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef __powerpc__
#define WORDS_BIGENDIAN
#endif

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysymdef.h>
#include <X11/keysym.h>
#include <X11/xpm.h>

#ifdef ZLIB
#include <zlib.h>
#endif

#include "xrmap.h"
#include "bitmaps.h"

#define MENU2 "Save map     RCfile       Cityfile     Mapfile      Quit"
#define MENU3 "File         Commands     Parameters   Fonts        "\
              "Continents   Categories   Colors/Features"
#define MENU4 "Configuration     US State lines    Rivers            "\
              "Inter. Boundar.   Coasts, Isl., Lakes"

static int spot_size[5]  = { 1, 2, 3, 4, 5};
static int zoom_value[5] = { 150, 1200, 3000, 6000, 10000};
static int popul_step[5] = { 3000000, 700000, 300000, 100000, 1};

static char city_filter0[] = "CR*";
static char loc_filter0[] = "BCIKLORSTV*";
static char *city_filter, *loc_filter;

static char *projtype[] = 
      { "Spherical", "Rectangular", "Mercator", "Miller" };
static char * itemtype[] = { "Categories", "Continents" };
static char * itemname[2][7] = {
      { "Config", "US", "Rivers", "International", "CIL", "", "" },
      { "", "Africa", "Asia", "Europe", "NAmerica", "SAmerica", "Oceanic" } };

static char *modes_descr[] = { 
          NULL, "command mode", "writing to file", "menu mode" };

static char *param_descr[] = {
  "Parameters", "projection", "width", "height", "zoom", "aspect", 
  "latitude", "longitude", "inclination", "slatitude", "slongitude", 
  "populsteps", "spotsizes", "thresholds", "cityfilter", "locfilter",
  "placename", "transparency",  };

static char *font_descr[] = {
  "Fonts", "font 0", "font 1", "font 2", "font 3", "font 4", "font 5" };

static char *contin_descr[] = { 
  "Continents", "Africa", "Asia", "Europe", "North America", "South America" };

static char *categ_descr[] = {
  "Categories", "US State lines", "Rivers", "International Boundaries", 
  "Coasts, Islands and Lakes" };

static char **feature_descr[5];

static int feature_length[5];
static int feature_index[5][17];

Display * dpy;
Visual * visual;
Window Root, win;
Colormap cmap = 0;
GC gc = 0, textgc = 0;
Atom wm_delete_window, wm_protocols;
Pixel black, white;
Pixmap textpix = 0;

XFontStruct * xfont[6];

int scr;
int color_depth;
int color_pad;
int bigendian;

int fontasc[6], fonth[6], fontasc[6], char_width;

int projection;
int numpt, printedpt, numcurves = 0;
int numtz, numloc;

int runlevel = 0;
int vectordata = 0;
int img_output = 0;
int do_menu = 0;
int reverse = 0;
int but_pressed = 0;
int line_clicked;
int overwrite = 1;
int change_made;
int point_shown = 0;
int color_change;
int caret_pos = 0;
int esc_width = 0;
int place_name = 0;
int box_x, box_y, box_width, box_height, box_spacing, box_numitems;
int cmd_mode = 0, cmd_type = 0;
int line_pos;
int city_shown;
int city_found;

char * cmd_string;

double earth_radius, aspect, coordconv, preview;

struct timezone **timezones = NULL;
struct location **locations = NULL;

/* The following subroutine is the only one borrowed from 
   the original mapping code -- strait forward bit flipping stuff
   which was contributed by Joe Dellinger (no email for me to
   contact him) 
   If you have a better way of accomplishing this, please
   Let me know as I'm sure this can't be super efficient
*/

#ifdef WORDS_BIGENDIAN
void bitflip (BIT32 *x) {
	BIT32 y=0;
	
	y |= 0xFF000000 & ((0x000000FF & *x) << 24);
	y |= 0x00FF0000 & ((0x0000FF00 & *x) << 8);
	y |= 0x0000FF00 & ((0x00FF0000 & *x) >> 8);
	y |= 0x000000FF & ((0xFF000000 & *x) >> 24);
	*x = y;
}
#endif /* WORDS_BIGENDIAN */

/* return a pointer to the color struct that 
   contains the rgb values for a color
   or null if nothing is found that matches
*/
void get_color (char* c, struct imageLayout * scene, int i, int j) {
        XColor color, exact;
	struct itemdata *item;
        int bool = 0;
        
	if (!c) return;
	item = &scene->item[i][j];
	if (*c == '+') {
	   item->bool = ON;
	   bool = 1;
	   ++c;
	}
	if (*c == '-') {
	   item->bool = OFF;
	   bool = 1;
	   ++c;
	}
	if (!(*c)) return;

	if (cmap) {
           if (XAllocNamedColor(dpy, cmap, c, &color, &exact)!=(Status)1)
   	      return;
   	   item->r = exact.red;
   	   item->g = exact.green;
   	   item->b = exact.blue;
           item->pix = color.pixel;
   	   if (reverse && (i>0 || j>=C_BG_MAP)) {
   	      if (color_depth <= 8) {
                 color.red = 65535 - exact.red;
                 color.green = 65535 - exact.green;
                 color.blue = 65535 - exact.blue;
                 XAllocColor(dpy, cmap, &color);
   	         item->pix = color.pixel;
   	      } else
   	         item->pix = ((1<<color_depth) - 1) - item->pix;
   	   }
	}

   	if (item->name) free(item->name);
	item->name = strdup(c);
}

void set_bg_fg(struct imageLayout * scene)
{
        if (win)
           XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_MAP].pix);
	if (gc) {
           XSetBackground(dpy, gc, scene->item[CONF][C_BG_TEXT].pix);
           XSetForeground(dpy, gc, scene->item[CONF][C_FG_TEXT].pix);
	}
}

void get_all_colors (struct imageLayout * scene)
{
        int i, j;
        char *col_name;

	if (color_depth<=8) {
	   if (cmap) XFreeColormap(dpy, cmap);
	   cmap = XCreateColormap(dpy, Root, visual, AllocNone);
	}

	for (i=0; i<5; i++)
 	    for (j=0; j<16; j++)
	       if (scene->item[i][j].bool != INACTIVE) {
		  col_name = strdup(scene->item[i][j].name);
		  if (i==CONF && j<C_BG_MAP)
		     scene->item[i][j].bool = 1;
	          get_color(col_name, scene, i, j);
		  free(col_name);
	       }

	if (color_depth<=8 && win)
           XSetWindowColormap(dpy, win, cmap);

	set_bg_fg(scene);
}

void set_opt_color(struct imageLayout * scene, char * string, int force)
{
	      int i, j;
	      char *ptr1, *ptr2;
	      char *buf;
	      
	      buf = (char *)malloc(strlen(string)+2);
              ptr1 = index(string, ',');
	      if (ptr1) {
		 strcpy(buf, string);
		 buf[ptr1-string] = '\0';
		 i = atoi(buf);
              } else {
		 free(buf);
		 return;
	      }
	      ++ptr1;
              ptr2 = index(ptr1, ',');
	      if (ptr2) {
		 strcpy(buf, ptr1);
		 buf[ptr2-ptr1] = '\0';
		 j = atoi(buf);
              } else {
		 free(buf);
		 return;
	      }
              ++ptr2;
	      if (*ptr2 && i>=0 && i<5 && j>=0 && j<16) {
		 if (scene->item[i][j].bool != INACTIVE || force) {
		    if (i==CONF && j<C_BG_MAP)
		       scene->item[i][j].bool = 1;
		    get_color(ptr2, scene, i, j);
		 }
	      }
	      free(buf);
}


int getWindowPlacement (win, x, y, w, h)
Window win;
int *x, *y;
unsigned int *w, *h;
{
   unsigned int b, d, n;
   Window root, parent, *children;

   XFlush(dpy);
   XQueryTree(dpy, win, &root, &parent, &children, &n);

   if (!parent) {
      fprintf(stderr, "Cannot query window tree!\n");
      return 0;
   }

   XGetGeometry(dpy, parent, &root, x, y, w, h, &b, &d);
   return 1;
}

void draw_separator(int x)
{
	XDrawLine(dpy, win, gc, x, 0, x, fonth[0]+5);
}

void drawWinString (struct imageLayout *scene, char *s, int position)
{
        int y0=0, yt=0, tw;

        if (position==TOP) {
	   y0 = fontasc[0] + 2;
	   yt = 0;
	}
        if (position==PRETOP) {
	   y0 = fontasc[0] + fonth[0]+6;
	   yt = fonth[0]+4;
	}
        if (position==PREPRETOP) {
	   y0 = fontasc[0] + 2*fonth[0]+10;
	   yt = 2*fonth[0]+8;
	}
        if (position==BOTTOM) {
	   y0 = scene->height - xfont[0]->max_bounds.descent - 4;
	   yt = scene->height - fonth[0] - 6;
	}
        if (position==PREBOTTOM) {
	   y0 = scene->height - xfont[0]->max_bounds.descent - fonth[0] - 8;
	   yt = scene->height - 2*fonth[0] - 10;
	}
        if (position==PREPREBOTTOM) {
	   y0 = scene->height - xfont[0]->max_bounds.descent - 2*fonth[0] - 12;
	   yt = scene->height - 3*fonth[0] - 14;
	}
        XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_TEXT].pix);
        XClearArea(dpy, win, 0, yt, scene->width, fonth[0]+6, False);
        XDrawImageString(dpy, win, gc, 4, y0, s, strlen(s));
	if (cmd_mode && position==BOTTOM) {
	   tw = XTextWidth(xfont[0], s, caret_pos);
           XSetWindowBackground(dpy, win, scene->item[CONF][C_CARET].pix);
           XClearArea(dpy, win, tw+4, yt+fonth[0]+3, char_width, 2, False);
           XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_TEXT].pix);
	}

	if (position==TOP) esc_width = 0;
	if (cmd_mode && position==TOP) {
           XSetWindowBackground(dpy, win, scene->item[CONF][C_ESCAPE].pix);
	   esc_width = XTextWidth(xfont[0], "Escape", 6)+16;
           XClearArea(dpy, win, scene->width-esc_width, 0, 
                                esc_width, fonth[0]+6, False);
	   draw_separator(scene->width-esc_width);
	   XSetBackground(dpy, gc, scene->item[CONF][C_ESCAPE].pix);
	   XDrawImageString(dpy, win, gc, scene->width-esc_width+8, fontasc[0]+2,
                            "Escape", 6);
           XSetBackground(dpy, gc, scene->item[CONF][C_BG_TEXT].pix);
	}

        XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_MAP].pix);
	XFlush(dpy);
}

int parse_cmd_line(struct imageLayout * scene, char *string);

void parse_rcfile (struct imageLayout *s) {
	FILE *f;
	int i,j,pos=0, commands=0;
	int ret, line_number=0;
	int cat, type;
	char string[256];
	char active[256];
	char color_name[256];
	char *comment=NULL;

	/* initialize stuff */
	bzero(string,256);

	/* read in the color config file and allocate colors */
	f = fopen(s->o.rcfile, "r");
	if (f == NULL) {
	   sprintf(string,"Couldn't open config rcfile `%s'", s->o.rcfile);
	   if (runlevel==0) {
	      fprintf(stderr, "%s\n", string);
	      exit(-2);
	   } else {
	      drawWinString(s, string, BOTTOM);
	      XFlush(dpy);
	      usleep(2000000);
	      return;
	   }
	}

	/* now parse the file */
	while ((ret = fread(string, 1, 255, f)) != 0) {
		line_number++;
		pos = (int) ftell(f) - ret;
		for (i=0; i<255; i++) {
			if (string[i] == '\n') {
				string[i] = '\0';
				fseek(f,pos+i+1, SEEK_SET);
				break;
			}
		}
		for (j=0; j<i; j++) {
			if (string[j] == '%')  {
				string[j] = '\0';
				break;
			}
		} 
		if (!strncasecmp(string, "<commandlines>", 9)) {
		   commands = 1;
		   continue;
		}
                if (commands==1) {
		   parse_cmd_line(s, string);
		   continue;
		}

		while (j>0) {
			j--;
			if (isspace(string[j])) {
				string[j] = '\0';
			} else {
				break;
			}
		}
		if (strlen(string) == 0) 
			continue;

		ret = sscanf(string, "%d %d %s %s", &cat, &type, active, color_name);
		comment = string;
		while (*comment && *comment!='!') ++comment;
		if (*comment=='!') ++comment;
		while (*comment && isspace(*comment)) ++comment;

		if (ret != 4 || !*comment) {
			fprintf(stderr,"Unable to parse the string [%s] on line %d in %s\n", 
				string,line_number,s->o.rcfile);	
			continue;
		}
		if (cat > 4 || type > 15 || cat < 0 || type < 0) {
			fprintf(stderr,"Declaration value out of range (%d,%d) on line %d in %s\n",
				cat,type, line_number, s->o.rcfile);
		} else {
			if (!strcasecmp(active, "ON"))
				s->item[cat][type].bool = ON;
			else
			if (!strcasecmp(active, "OFF"))
				s->item[cat][type].bool = OFF;
                  	if (s->item[cat][type].bool != INACTIVE) {
			        if (s->item[cat][type].name)
                                   free(s->item[cat][type].name);
			        if (s->item[cat][type].comment)
                                   free(s->item[cat][type].comment);
				s->item[cat][type].name = strdup(color_name);
				s->item[cat][type].comment = strdup(comment);
			}
  		}
	}
	fclose(f);
}

double angle_inc(double z)
{
        double norm[] = { 1.0, 0.75, 0.5, 0.3, 0.2, 0.15 };
        double val, valn, fact;
	int i;

	if (z<=1.0) return 15.0;

	val = 15.0/z;
	valn = 10.0;
	fact = 10.0;
	i=0;
	while (valn>val) {
	   ++i;
	   if (i>5) {
	      fact *= 0.1;
	      i = 0;
	   }
	   valn = norm[i]*fact;
	}

        return valn;
}

void check_spacing(struct imageLayout * scene)
{
        if (scene->o.spacing_lat< 1.0/scene->o.zoom)
	    scene->o.spacing_lat = angle_inc(scene->o.zoom);

        if (scene->o.spacing_lon< 1.0/scene->o.zoom)
	    scene->o.spacing_lon = angle_inc(scene->o.zoom);

}

void show_title(Window w)
{
        char title[80];
	int i;

	if (cmd_mode<3)
	   i = cmd_mode;
	else
	   i = 3;
	modes_descr[0] = projtype[projection];
	sprintf(title, "%s / %s", PACKAGE, modes_descr[i]);
        XStoreName(dpy, w, title);
}

void set_radius(struct imageLayout * scene) 
{
        earth_radius = (double)(scene->width/2);

        if (projection==SPHERICAL && scene->width>scene->height)
	   earth_radius = (double)(scene->height/2);
        coordconv = 180.0/earth_radius;
}

void set_font_params(int which)
{
        if (which<0 || which>5) return;
	fontasc[which] = xfont[which]->max_bounds.ascent;
	fonth[which] = fontasc[which] + xfont[which]->max_bounds.descent;
	if (which==0) char_width = XTextWidth(xfont[which], "w", 1);
}

int strpcmp(char *s, char *t, int n)
{
        int mode;

        if (n<0) {
	   mode = 0;
	   n = -n;
	} else
	   mode = 1;
        if (!s || !t) return -1;
	if (runlevel==0 && *s!='-' && *s!='+') return -1;

        if (*t!='+' && *t!='-') { 
           while(*s=='-') ++s; 
	}
	if (strlen(s)<n || strlen(s)>strlen(t)) return -1;
        n = strlen(s);

	if (mode) 
           return strncmp(s, t, n);
	else
           return strncasecmp(s, t, n);
}

int itemnumber(int mode, char * str)
{
       int i, run;
       char *endptr;

       if (!str) return -1;
       run = runlevel;
       runlevel = 1;
       for (i = 0; i <= 4+2*mode; i++)
           if (!strpcmp(str, itemname[mode][i], -2)) {
	      runlevel = run;
	      return i;
	   }
       runlevel = run;
       i = strtol(str, &endptr, 10);
       if (!endptr) return -1;
       return i;
}

int set_flag(struct imageLayout * scene, char *c, int mode)
{
    char * number, *ptr1, *ptr2;
    char str[256];
    int i, tot;

    if (!strcasecmp(c, "all")) {
       if (mode) 
          scene->o.desired_continents = 0xFFFFFFFF;
       else
          scene->o.desired_categories = 0xFFFFFFFF;
       return 0;
    }
    
    number = (char *) malloc(strlen(c)+2);
    ptr1 = c;
    tot = 0;

    while (ptr1 && *ptr1) {
       ptr2 = index(ptr1, ',');
       strcpy(number, ptr1);
       if (ptr2) number[ptr2-ptr1] = '\0';
       i = itemnumber(mode, number);
       if (i < 1 || i > 4+2*mode) {
	  sprintf(str,"%s must be in the range of 1-%d\n",
		  itemtype[mode], 4+2*mode);
	  if (runlevel)
             drawWinString(scene, str, BOTTOM);
	  free(number);
	  return 1;
       }
       tot |= (0x00000001 << (i-1));
       if (ptr2)
          ptr1 = ptr2+1;
       else
	  ptr1 = NULL;
    }
    if (mode) 
       scene->o.desired_continents = tot;
    else
       scene->o.desired_categories = tot;
    free(number);
    return 0;
}

void usage()
{
	printf("Usage: xrmap [-projection TYPE] [-zoom VALUE] [-aspect VALUE]\n");
	printf("             [-latitude VALUE] [-longitude VALUE] [-inclination VALUE]\n");
	printf("             [-mapfile NAME] [-datafile NAME] [-rcfile NAME] [-outfile NAME]\n");
        printf("             [-continent LIST] [-category LIST]\n");
        printf("             [-img_output] [-vectordata SPACING] [-preview TIME(sec)]\n");
	printf("\n");	
	printf("\t-projection <type>        (0=Spherical, 1=Rectangular,\n\t\t\t\t   2=Mercator,  3=Miller)\n");
	printf("\t-aspect <value>           Aspect = ratio height/width (proj 1,2,3)\n");
	printf("\t-zoom <value>             Zoom Value, <= 1.0 for whole Earth\n");
	printf("\t-width <value>            The width of the generated image\n");
	printf("\t-height <value>           The height of the generated image\n");
	printf("\t-latitude <LAT value>     Rotates the Earth to LAT value\n");
	printf("\t-longitude <LON value>    Rotates the Earth to LON value\n");
	printf("\t-inclination <INC value>  Shows the Equator with inclination INC\n");
	printf("\t-spacing <value>          Spacing between gridlines\n");
	printf("\t-slatitude <value>        Spacing between parallels\n");
	printf("\t-slongitude <value>       Spacing between meridians\n");
	printf("\t-spotsizes <values>       Size of city spots, = 0..5\n");
	printf("\t-populsteps <values>      Population steps for spot sizes 1..5\n");
	printf("\t-thresholds <values>      Thresholds for showing cities\n");
	printf("\t-rcfile <path_to_file>    Name of the private configuration file\n");
	printf("\t-mapfile <path_to_file>   File containing the CIA vector map\n");
	printf("\t-datafile <path_to_file>  File containing timezones/cities/locations\n");
	printf("\t-outfile <path_to_file>   File to which image output is to be written\n");
	printf("\t-ppmoutput                Create a .ppm file and exit xrmap (no GUI)\n");
	printf("\t-xpmoutput                Create a .xpm file and exit xrmap (no GUI)\n");
	/*	printf("\t-psoutput                 Create a .ps file and exit xrmap (no GUI)\n"); */
	printf("\t-vectordata <number>      Write vector data spaced by <number> (no GUI)\n");
	printf("\t-preview <time>           Preview <time> before output and exit\n");
	printf("\t-continent <list>         `all' or comma separated list of items:\n");
	printf("\t  1=(Af)rica, 2=(As)ia, 3=(Eu)rope, 4=N.America (NA), 5=S.America (SA)\n");
	printf("\t-category <list>          `all' or comma separated list of items:\n");
	printf("\t  1=(US) State Boundaries,       2=(Riv)ers,\n");
	printf("\t  3=(Int)ernational Boundaries,  4=Coasts, Islands, Lakes (CIL)\n");
	printf("\t-item <cat,i,()color>    Set flag() and color of item[cat][i], 0<i<16\n");
	printf("\t-font <i,string>          Set font[i] as defined by string (menu i=0)\n");
	printf("\t-reverse                  Reverse colors at startup\n");
	printf("\t-resetcolors              Reset colors (run mode only)\n");
	printf("\tplacename                Vary/don't vary name position\n");
	printf("\tcities                   Display/don't display cities\n");
	printf("\tlocations                Display/don't display other locations\n");
	printf("\tgridlines                Display/don't display grid lines\n");
	printf("\tearthcontour             Display/don't display Earth contour\n");
	printf("\ttransparency             Make/don't make the Earth transparent\n");
	printf("\t-h                        This screen\n");	
	printf("\t-v                        Version information\n");	
}

void set_proj(char *s)
{
        int run;
	run = runlevel;
	runlevel = 1;
	if (!strpcmp(s, "spherical", -3))
           projection = SPHERICAL;
	else
	if (!strpcmp(s, "rectangular", -3))
	   projection = RECTANGULAR;
	else
	if (!strpcmp(s, "mercator", -3))
	   projection = MERCATOR;
	else
	if (!strpcmp(s, "miller", -3))
	   projection = MILLER;
	else {
           projection = atoi(s); 
           if (projection<SPHERICAL) projection = SPHERICAL;
	   if (projection>MILLER) projection = MILLER;
	}
	runlevel = run;
}

void create_textpix(int width)
{
        int i, mfonth;
        if (textpix)
	   XFreePixmap(dpy, textpix);
	mfonth = 0;
	for (i=1; i<=5; i++)
	    if (fonth[i]>mfonth) mfonth = fonth[i];
	textpix = XCreatePixmap(dpy, Root, width, mfonth, 1);
}

void set_font(char *s, int which)
{
	XFontStruct * nfont;
	nfont = XLoadQueryFont(dpy, s);
	if (nfont == (XFontStruct *)NULL) return;
        if (xfont[which]) XFreeFont(dpy, xfont[which]);
	xfont[which] = nfont;
        if (which==0)
           XSetFont(dpy, gc, nfont->fid);
        if (textpix)
	   XFreePixmap(dpy, textpix);
	textpix = 0;
	set_font_params(which);
}

void putPicture (struct imageLayout *s) {
     XPutImage(dpy, win, gc,
                    s->im, 0, 0, 0, 0,
                    s->im->width, s->im->height);
     line_pos = -1;
     city_found = -1;
     city_shown = 0;
     point_shown = 0;
     XFlush(dpy);
}

void freeTZ()
{
int i;
   if (timezones) { 
      for (i=0; i<numtz; i++) {
         free(timezones[i]->code);
         free(timezones[i]->flags);
         free(timezones[i]->name);
         free(timezones[i]->tzvalue);
	 free(timezones[i]);
      }
      free(timezones);
      timezones = NULL;
   }
   numtz = 0;
}

void freeloc()
{
int i;
   if (locations) {
      for (i=0; i<numloc; i++) {
         free(locations[i]->name);
         free(locations[i]->region);
	 free(locations[i]);
      }
      free(locations);
      locations = NULL;
   }
   numloc = 0;
}

int round2int (double x)
{
        if (x>0) 
           return (int)(x+0.5) ;
	else
	   return -(int)(-x+0.5);
}

enum { T_NONE=0, T_TIMEZONES, T_CITIES, T_LOCATIONS};

void parse_datafile (struct imageLayout *s) {
	FILE *f;	
	char buf[512];
        char field[11][128];
	char *sect[] = { "", "timezones", "cities" };
        char *str, *ptr, *ptrp;
	int bool, i, numfields, datatype;

	if (!s->o.datafile || !*s->o.datafile) return;

	f = fopen(s->o.datafile, "r");
	if (f == NULL) {
	   sprintf(buf,"Couldn't open datafile `%s'", s->o.datafile);
	   if (runlevel==0) {
	      fprintf(stderr, "%s\n", buf);
	      exit(-1);
	   } else {
	      drawWinString(s, buf, BOTTOM);
	      XFlush(dpy);
	      usleep(2000000);
	      return;
	   }
	}

	numfields = 0;
	datatype = T_NONE;
        while((str=fgets(buf, 510, f))) {
	   while (*str && isspace(*str)) ++str;
	   if (!*str || *str=='%') continue;
	   if (!strncmp(str, "<reset_TZ>", 10)) {
              freeTZ();
	      continue;
	   }
	   if (!strncmp(str, "<reset_cities>", 14)) {
              freeloc();
	      continue;
           }
	   if (!strncmp(str, "<timezones>", 11)) {
	      datatype = T_TIMEZONES;
	      numfields = 5;
	      continue;
	   }
	   if (!strncmp(str, "<cities>", 8)) {
	      datatype = T_CITIES;
	      numfields = 11;
	      continue;
	   }
	   if (!strncmp(str, "<locations>", 11)) {
	      datatype = T_LOCATIONS;
	      numfields = 7;
	      continue;
	   }
	   if (datatype==T_TIMEZONES && !strncmp(str, "</timezones>", 12)) {
	      datatype = T_NONE;
	      numfields = 0;
	      continue;
	   }
	   if (datatype==T_CITIES && !strncmp(str, "</cities>", 9)) {
	      datatype = T_NONE;
	      numfields = 0;
	      continue;
	   }
	   if (datatype==T_LOCATIONS && !strncmp(str, "</locations>", 12)) {
	      datatype = T_NONE;
	      numfields = 0;
	      continue;
	   }
           bool = 0;
           for (i=0; i<numfields; i++) {
              while (*str && isspace(*str)) ++str;
	      ptr = str;
              while (*ptr && *ptr!='|') ++ptr;
	      if (!*ptr && i<numfields-1) bool=1;
              *ptr = '\0';
              ptrp = ptr+1;
	      --ptr;
              while(isspace(*ptr)) { *ptr = '\0'; --ptr; }
              strcpy(field[i], str);
	      str = ptrp;
           }
           if (bool)
              fprintf(stderr, "%s : field missing in %s section !!\n", 
	         field[0], sect[datatype]);
	   if (datatype==T_TIMEZONES) {
              timezones = (struct timezone **)
                  realloc(timezones, (numtz+1)*sizeof(struct timezone *));
	      timezones[numtz] = malloc(sizeof(struct timezone));
	      timezones[numtz]->code = strdup(field[0]);
	      timezones[numtz]->flags = strdup(field[2]);
	      timezones[numtz]->name = strdup(field[3]);
	      timezones[numtz]->tzvalue = strdup(field[4]);
	      if (!strcasecmp(field[1], "AF")) timezones[numtz]->continent = 0;
	      if (!strcasecmp(field[1], "AS")) timezones[numtz]->continent = 1;
	      if (!strcasecmp(field[1], "EU")) timezones[numtz]->continent = 2;
	      if (!strcasecmp(field[1], "NA")) timezones[numtz]->continent = 3;
	      if (!strcasecmp(field[1], "SA")) timezones[numtz]->continent = 4;
	      ++numtz;
	   }
           if (datatype==T_CITIES) {
              locations = (struct location **)
                  realloc(locations, (numloc+1)*sizeof(struct location *));
	      locations[numloc] = malloc(sizeof(struct location));
	      locations[numloc]->obj = 'c';
	      locations[numloc]->country = strdup(field[0]);
	      locations[numloc]->name = strdup(field[1]);
	      locations[numloc]->region = strdup(field[3]);
	      locations[numloc]->status = field[2][0];
	      locations[numloc]->rank = atoi(field[4]);
	      locations[numloc]->worldrank = atoi(field[5]);
	      locations[numloc]->population = atoi(field[6]);
	      locations[numloc]->lat = atof(field[7]);
	      locations[numloc]->lon = atof(field[8]);
	      if (index(field[9],'?'))
	         locations[numloc]->alt = -30000;
	      else
	         locations[numloc]->alt = round2int(atof(field[9]));
	      locations[numloc]->tz = -1;
              for (i=0; i<numtz; i++) {
		 if (!strcmp(field[10], timezones[i]->code)) {
		    locations[numloc]->tz = i;
                    break;
		 }
	      }
	      ++numloc;
	   }
           if (datatype==T_LOCATIONS) {
              locations = (struct location **)
                  realloc(locations, (numloc+1)*sizeof(struct location *));
	      locations[numloc] = malloc(sizeof(struct location));
	      locations[numloc]->obj = 'l';
	      locations[numloc]->status = field[2][0];
	      locations[numloc]->name = strdup(field[0]);
	      locations[numloc]->country = NULL;
	      locations[numloc]->region = NULL;
	      locations[numloc]->rank = (int)(field[4][1]-'0');
	      locations[numloc]->worldrank = 0;
	      locations[numloc]->population = atoi(field[3]);
	      locations[numloc]->lat = atof(field[5]);
	      locations[numloc]->lon = atof(field[6]);
	      locations[numloc]->alt = 0;
              locations[numloc]->tz = -1;
	      if (!strcasecmp(field[1], "AF")) locations[numloc]->tz = 1;
	      else
	      if (!strcasecmp(field[1], "AS")) locations[numloc]->tz = 2;
	      else
	      if (!strcasecmp(field[1], "EU")) locations[numloc]->tz = 3;
	      else
	      if (!strcasecmp(field[1], "NA")) locations[numloc]->tz = 4;
	      else
	      if (!strcasecmp(field[1], "SA")) locations[numloc]->tz = 5;
	      else
	      if (!strcasecmp(field[1], "OC")) locations[numloc]->tz = 6;
	      ++numloc;
	   }
	}
	      
	fclose(f);
}

void load_city_filter(char *str)
{
     char ch;
     int j, k;

     *city_filter ='\0';
     k = 0;
     for (j=0;j<strlen(str);j++) {
        ch = toupper(str[j]);
        if (index(city_filter, ch)) continue;
	if (index(city_filter0, ch)) {
	   city_filter[k] = ch;
	   ++k;
	   city_filter[k] = '\0';
	}
     }
}

void load_loc_filter(char *str)
{
     char ch;
     int j, k;

     *loc_filter ='\0';
     k = 0;
     for (j=0;j<strlen(str);j++) {
        ch = toupper(str[j]);
        if (index(loc_filter, ch)) continue;
	if (index(loc_filter0, ch)) {
	   loc_filter[k] = ch;
	   ++k;
	   loc_filter[k] = '\0';
	}
     }
}
		
int parse_cmd_line(struct imageLayout * scene, char *str)
{
        char *string, *ptr, *ptr_err;
	int w0, h0;

	if (!str || !*str) return -1;
	string = str;

	w0 = scene->width;
	h0 = scene->height;

	while ((ptr = (char*) strtok(string, " ")) != NULL) {
	   ptr_err = ptr;
	   string = NULL;
	   if (!runlevel && !strpcmp(ptr, "help", 1)) {
	      usage();
	      exit(0);
	   } else
	   if (!runlevel && !strpcmp(ptr, "version", 1)) {
              printf("%s version %s (%s)\n", PACKAGE, VERSION, DATEPKG);
              exit(0);
	   }
           else
	   if (!runlevel && !strpcmp(ptr, "menu", 3)) {
              do_menu = 1;
	   }
           else
	   if (!strpcmp(ptr, "mapfile", 4)) {
	      if (scene->o.mapfile) free(scene->o.mapfile);
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
                 scene->o.mapfile = strdup(ptr);
	      else
		 goto error;
	   } 
           else 
	   if (!strpcmp(ptr, "rcfile", 3)) {
	      if (scene->o.rcfile) free(scene->o.rcfile);
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr || (ptr && !strcasecmp(ptr, "NULL")))
		 scene->o.rcfile = NULL;
	      else
                 scene->o.rcfile = strdup(ptr);
	      if (runlevel) parse_rcfile(scene);
	   } 
           else
	   if (!strpcmp(ptr, "datafile", 5)) {
	      if (scene->o.datafile) free(scene->o.datafile);
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr || (ptr && !strcasecmp(ptr, "NULL")))
		 scene->o.datafile = NULL;
	      else
                 scene->o.datafile = strdup(ptr);
	      if (runlevel) parse_datafile(scene);
	   }
           else
	   if (!strpcmp(ptr, "outfile", 4)) {
	      if (scene->o.outfile) free(scene->o.outfile);
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr || (ptr && !strcasecmp(ptr, "NULL")))
		 scene->o.outfile = NULL;
	      else
                 scene->o.outfile = strdup(ptr);
	   }
	   else
	   if (!runlevel && !strpcmp(ptr, "vectordata", 3)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) vectordata = atoi(ptr); else goto error;
              if (vectordata==0) goto error;
	      if (img_output) goto error;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "ppmoutput", 3)) {
              img_output = 1;
	      if (vectordata) goto error;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "xpmoutput", 3)) {
              img_output = 2;
	      if (vectordata) goto error;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "psoutput", 2)) {
              img_output = 3;
	      if (vectordata) goto error;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "preview", 7)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) preview = atof(ptr); else goto error;
              if (preview<0.0) goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "longitude", 2)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.yrot = atof(ptr); else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "latitude", 2)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.xrot = atof(ptr); else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "inclination", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.zrot = atof(ptr); else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "reverse", 1)) {
    	      reverse = 1 - reverse;
	      if (runlevel) get_all_colors(scene);
	   } 
	   else
	   if (!strpcmp(ptr, "resetcolors", 3)) {
	      if (runlevel && color_depth<=8) get_all_colors(scene);
	   } 
           else
	   if (!strpcmp(ptr, "font", 2)) {
	      int num;
	      char *ptr2;
              ptr = (char*) strtok(NULL, " ");
      	      if (!ptr) goto error;
	      if (!(ptr2=index(ptr,','))) goto error;
	      *ptr2 = '\0';
              num = atoi(ptr);
	      *ptr2 = ',';
	      if (num<0 || num>5) goto error;
	      if (scene->o.font[num]) free(scene->o.font[num]);
	      scene->o.font[num] = strdup(ptr2+1);
	      if (!runlevel) continue;
	      set_font(ptr2+1, num);
	   }
	   else
	   if (!strpcmp(ptr, "width", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->width = atoi(ptr); else goto error;
	      if (scene->width < 16) goto error;
              if (runlevel && textpix) {
	         XFreePixmap(dpy, textpix);
		 textpix = 0;
	      }
	   }
	   else
	   if (!strpcmp(ptr, "height", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->height = atoi(ptr); else goto error;
	      if (scene->height < 16) goto error;
	   }
	   else
	   if (!strpcmp(ptr, "item", 2)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) set_opt_color(scene, ptr, 0); else goto error;
	      set_bg_fg(scene);
	   }
	   else
	   if (!strpcmp(ptr, "category", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
                 if (set_flag(scene, ptr, 0))
		    goto error;
	      }
              else 
                 goto error;
	   }
	   else
	   if (!strpcmp(ptr, "continent", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
		 if (set_flag(scene, ptr, 1))
		    goto error;
	      }
              else 
                 goto error;
	   }
	   else
	   if (!strpcmp(ptr, "projection", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) set_proj(ptr); else goto error;
              if (runlevel) show_title(win);
	   }
	   else
	   if (!strpcmp(ptr, "aspect", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) aspect = atof(ptr); else goto error;
	      if (aspect<0.01) goto error;
	   }
	   else
	   if (!strpcmp(ptr, "zoom", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.zoom = atof(ptr); else goto error;
	      if (scene->o.zoom<0.01) goto error;
	   }
	   else
	   if (!strpcmp(ptr, "slatitude", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.spacing_lat = atof(ptr); else goto error;
	      check_spacing(scene);
	   }
	   else
	   if (!strpcmp(ptr,"slongitude", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.spacing_lon = atof(ptr); else goto error;
	      check_spacing(scene);
	   }
	   else
	   if (!strpcmp(ptr, "spacing", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->o.spacing_lat = atof(ptr); else goto error;
	      check_spacing(scene);
	      scene->o.spacing_lat = scene->o.spacing_lon;
	   } 
	   else
	   if (!strpcmp(ptr, "spotsizes", 3)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &spot_size[0], &spot_size[1], &spot_size[2],
                            &spot_size[3], &spot_size[4]))<5)
		 for (j=i; j<5; j++) spot_size[j]=spot_size[i-1];
	   }                          
	   else
	   if (!strpcmp(ptr, "populsteps", 3)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &popul_step[0], &popul_step[1], &popul_step[2],
                            &popul_step[3], &popul_step[4]))<5)
		 for (j=i; j<5; j++) popul_step[j]=popul_step[i-1];
	   }                          
	   else
	   if (!strpcmp(ptr, "thresholds", 3)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &zoom_value[0], &zoom_value[1], &zoom_value[2],
                            &zoom_value[3], &zoom_value[4]))<5)
		 for (j=i; j<5; j++) zoom_value[j]=zoom_value[i-1];
	   }                          
           else
	   if (!strpcmp(ptr, "cityfilter", 5)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
		 load_city_filter(ptr);
              else goto error;
	   }
           else
	   if (!strpcmp(ptr, "locfilter", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
		 load_loc_filter(ptr);
              else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "+placename", 3)) 
	      place_name = 1;
           else
	   if (!strpcmp(ptr, "-placename", 3)) 
	      place_name = 0;
           else
	   if (!strpcmp(ptr, "+cities", 2)) {
	      scene->item[CONF][C_CITY_CIRCLE].bool = ON;
	      scene->item[CONF][C_CITY_TEXT].bool = ON;
	   }
	   else
	   if (!strpcmp(ptr, "-cities", 2)) {
	      scene->item[CONF][C_CITY_CIRCLE].bool = OFF;
	      scene->item[CONF][C_CITY_TEXT].bool = OFF;
	   }
	   else
	   if (!strpcmp(ptr, "+earthcontour", 2)) 
              scene->item[CONF][C_EARTH_CONTOUR].bool = ON;
	   else
	   if (!strpcmp(ptr, "-earthcontour", 2)) 
              scene->item[CONF][C_EARTH_CONTOUR].bool = OFF;
	   else
	   if (!strpcmp(ptr, "+locations", 2)) {
              scene->item[CONF][C_LOC_LAND].bool = ON;
              scene->item[CONF][C_LOC_WATER].bool = ON;
	   } else
	   if (!strpcmp(ptr, "-locations", 2)) {
              scene->item[CONF][C_LOC_LAND].bool = OFF;
              scene->item[CONF][C_LOC_WATER].bool = OFF;
	   } 
	   else
	   if (!strpcmp(ptr, "+transparency", 2)) 
              scene->transparent = ON;
	   else
	   if (!strpcmp(ptr, "-transparency", 2)) 
              scene->transparent = OFF;
	   else
	   if (!strpcmp(ptr, "+gridlines", 2)) {
              scene->item[CONF][C_LAT_GRID].bool = ON;
              scene->item[CONF][C_LON_GRID].bool = ON;
	   }
	   else
	   if (!strpcmp(ptr, "-gridlines", 2)) {
              scene->item[CONF][C_LAT_GRID].bool = OFF;
              scene->item[CONF][C_LON_GRID].bool = OFF;
	   }
	   else {
	      char *str;
	      ptr_err = ptr;
	   error:
	      str = (char *)malloc(strlen(ptr_err)+100);
              sprintf(str, "%s : option invalide. "
                       "Try `xrmap -h' for more information", ptr_err);
	      *cmd_string = '\0';
	      caret_pos = 0;
	      if (runlevel) {
	         drawWinString(scene, str, BOTTOM);
	         usleep(2000000);
	      } else
		 fprintf(stderr, "%s\n", str);
	      free(str);
	      return 1;
	   }
	}
	if (runlevel) {
	   if (scene->width<=0) scene->width = w0;
	   if (scene->height<=0) scene->height = h0;
	   if (scene->width!=w0 || scene->height!=h0) { 
	      set_radius(scene);
              XResizeWindow(dpy, win, scene->width, scene->height);
	   }
           if (textpix) {
               XFreePixmap(dpy, textpix);
	       textpix = 0;
	   }
	}
	return 0;
}

void value2str(struct imageLayout * scene, int i, char * str, int mode)
{
        int j;
	char c;

        *str = '\0';   
	if (cmd_mode==4) {
	   if (cmd_type==1) {
	      switch(i) {
		 case 1:  if (mode) 
                             sprintf(str, "%d = %s", projection,
                                  projtype[projection]); 
                          else
                             sprintf(str, "%d", projection);
                          break;
		 case 2:  sprintf(str, "%d", scene->width); break;
		 case 3:  sprintf(str, "%d", scene->height); break;
		 case 4:  sprintf(str, "%.3f", scene->o.zoom); break;
		 case 5:  sprintf(str, "%.3f", aspect); break;
		 case 6:  sprintf(str, "%.3f", scene->o.xrot); break;
		 case 7:  sprintf(str, "%.3f", scene->o.yrot); break;
		 case 8:  sprintf(str, "%.3f", scene->o.zrot); break;
		 case 9:  sprintf(str, "%.3f", scene->o.spacing_lat); break;
		 case 10: sprintf(str, "%.3f", scene->o.spacing_lon); break;
		 case 11: sprintf(str, "%d,%d,%d,%d,%d", 
			    popul_step[0], popul_step[1], popul_step[2],
                            popul_step[3], popul_step[4]); 
                          break;
		 case 12: sprintf(str, "%d,%d,%d,%d,%d", 
			    spot_size[0], spot_size[1], spot_size[2],
                            spot_size[3], spot_size[4]); 
                          break;
		 case 13: sprintf(str, "%d,%d,%d,%d,%d", 
			    zoom_value[0], zoom_value[1], zoom_value[2],
                            zoom_value[3], zoom_value[4]); 
                          break;
	         case 14: strcpy(str, city_filter);
		          break;
	         case 15: strcpy(str, loc_filter);
		          break;
	         case 16: if (place_name) 
                             strcpy(str, "+"); 
                          else 
                             strcpy(str, "-");
		          break;
	         case 17: if (scene->transparent)
                             strcpy(str, "+"); 
                          else 
                             strcpy(str, "-");
		          break;
	         default: break;
	      }
	   }

	   if (cmd_type==2) {
              if (i>=1 && i<=6)
	         strcpy(str, scene->o.font[i-1]);
	   }

	   if (cmd_type==3) {
	      if ((1<<(i-1)) & scene->o.desired_continents)
		 strcpy(str, "+");
	      else
		 strcpy(str, "-");
	   }
	   if (cmd_type==4) {
	      if ((1<<(i-1)) & scene->o.desired_categories)
		 strcpy(str, "+");
	      else
		 strcpy(str, "-");
	   }
	}
        if (cmd_mode==5) {
	   j = feature_index[cmd_type-1][i];
	   c = (scene->item[cmd_type-1][j].bool)? '+' : '-';
	   sprintf(str, "%c%s", c, scene->item[cmd_type-1][j].name);
        }
}

void str2value(struct imageLayout * scene, int i, char * str)
{
        int j;
	double f;

	if (cmd_mode==4) {
	   if (cmd_type==1) {
	      switch(i) {
		 case 1:  set_proj(str); break;
		 case 2:  j = atoi(str); 
		          if (j<16) break;
			  scene->width = j;
			  XResizeWindow(dpy, win, scene->width, scene->height);
                          if (textpix) {
                             XFreePixmap(dpy, textpix);
	                     textpix = 0;
	                  }
		          break;
		 case 3:  j = atoi(str); 
		          if (j<16) break;
			  scene->height = j; 
			  XResizeWindow(dpy, win, scene->width, scene->height);
                          break;
		 case 4:  f = atof(str); 
		          if (f<0.01) break;
			  scene->o.zoom = f;
		          break;
		 case 5:  f = atof(str); 
		          if (f<0.01) break;
                          aspect = f; 
                          break;
		 case 6:  scene->o.xrot = atof(str); break;
		 case 7:  scene->o.yrot = atof(str); break;
		 case 8:  scene->o.zrot = atof(str); break;
		 case 9:  scene->o.spacing_lat = atof(str); break;
		 case 10: scene->o.spacing_lon = atof(str); break;
		 case 11: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &popul_step[0], &popul_step[1], &popul_step[2],
                            &popul_step[3], &popul_step[4]))<5)
		             while(j<5) { popul_step[j]=popul_step[j-1]; ++j; }
                          break;
		 case 12: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &spot_size[0], &spot_size[1], &spot_size[2],
                            &spot_size[3], &spot_size[4]))<5)
		             while(j<5) { spot_size[j]=spot_size[j-1]; ++j; }
                          break;
		 case 13: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &zoom_value[0], &zoom_value[1], &zoom_value[2],
                            &zoom_value[3], &zoom_value[4]))<5)
		             while(j<5) { zoom_value[j]=zoom_value[j-1]; ++j; }
                          break;
	         case 14: load_city_filter(str);
			  break;
	         case 15: load_loc_filter(str);
			  break;
	         case 16: if (*str == '+') 
		             place_name = 1;
		          else
		             place_name = 0;
                          break;
	         case 17: if (*str == '+') 
		             scene->transparent = 1;
		          else
		             scene->transparent = 0;
                          break;
		 default: break;
	      }
	   }
	   if (cmd_type==2) { 
              if (i>=1 && i<=6) {
		 i -= 1;
	         if (scene->o.font[i]) free(scene->o.font[i]); 
		 scene->o.font[i] = strdup(str); 
	         set_font(scene->o.font[i], i);
		 putPicture(scene);
	      }
	   }

	   if (cmd_type==3) {
              j = 1<<(i-1);
	      if (!strcmp(str, "+")) scene->o.desired_continents |= j;
	      if (!strcmp(str, "-")) scene->o.desired_continents &= ~j;
	   }
	   if (cmd_type==4) {
              j = 1<<(i-1);
	      if (!strcmp(str, "+")) scene->o.desired_categories |= j;
	      if (!strcmp(str, "-")) scene->o.desired_categories &= ~j;
	   }
	}

        if (cmd_mode==5) {
	   j = feature_index[cmd_type-1][i];
	   if (cmd_type==1 && j<C_BG_MAP)
	      scene->item[CONF][j].bool = 1;	      
           get_color(str, scene, cmd_type-1, j);
	   color_change = 1;
	   if (cmd_type==1 && j<C_BG_MAP) return;
        }
        change_made = 1;
}

void drawRectCmd(struct imageLayout * scene, char **paramstr, int n)
{
        int i, m;
	char str[256];

	drawWinString(scene, paramstr[0], TOP);
	box_width = 0;
	box_height = fonth[0]+6;
	box_spacing = fonth[0]+10;
	box_numitems = n;
	box_x = 6;
	box_y = (7*box_spacing)/5;
        XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_TEXT].pix);
        XClearArea(dpy, win, 0, box_y+box_spacing/2-1, 
                   scene->width, (n+1)*box_spacing, False);
        for (i=1; i<=n; i++) {
	   XDrawImageString(dpy, win, gc, 
                            box_x+6, box_y+i*box_spacing+fontasc[0]+3,
	                    paramstr[i], strlen(paramstr[i]));
           m = XTextWidth(xfont[0], paramstr[i], strlen(paramstr[i]))+12;
	   if (m>box_width) box_width = m;
	}
        for (i=1; i<=n; i++) {
	   XDrawRectangle(dpy, win, gc, box_x, box_y+i*box_spacing,
		               box_width, box_height);
	   if (line_clicked==i)
	      XDrawRectangle(dpy, win, gc, box_x+1, box_y+i*box_spacing+1,
		               box_width-2, box_height-2);
	   value2str(scene, i, str, 1);
	   if (*str) {
	      XDrawImageString(dpy, win, gc, 
                               box_x+box_width+5*char_width, 
                               box_y+i*box_spacing+fontasc[0]+3,
	                       str, strlen(str));

	   }
	}
        XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_MAP].pix);
	XFlush(dpy);
}

void drawWinCmd(struct imageLayout * scene)
{
        int i, j;
	char str[256];
        if (cmd_mode==1) {
           strcpy(str,
                "Switches | options: "
                "<Enter> c e g l t   "
                "p w h z a lat lon inc s slat slon font item cat con");
           drawWinString(scene, str, TOP);
	   j = XTextWidth(xfont[0], str, 44);
	   draw_separator(j);
	}
        if (cmd_mode==2) {
	   if (cmd_type==0) {
	      strcpy(str, MENU2);
	      drawWinString(scene, str, TOP);
	      for (i=1; i<=4; i++) {
	          j = XTextWidth(xfont[0], str, 13*i-1);
	          draw_separator(j);
	      }
	   }
           if (cmd_type==1) {
   	      if (overwrite==0)
   	         sprintf(str, "Writing to file: specify file name at bottom "
                              "(*.ppm , *.xpm or *.ps file)");
   	      else
   	      if (overwrite==1)
   	         sprintf(str, "Writing: file %s(.gz) already exists. "
                     "Overwrite (<Enter>) ? Don't (<Esc>) ?", 
                     scene->o.outfile);
              else 
   	      if (overwrite==-1) {
                 #ifdef ZLIB
   	         char compl[6];
   	         if (scene->o.outfile &&
                     strcmp(scene->o.outfile+strlen(scene->o.outfile)-3,".gz"))
   		    strcpy(compl, ".gz");
   	         else
   		    *compl = '\0';
   	         sprintf(str, "Writing image to file %s%s ...",
   	                 scene->o.outfile, compl);
                 #else
   	         sprintf(str, "Writing image to file %s ...",
   	   	         scene->o.outfile);
                 #endif
   	      }
	   }
           if (cmd_type==2)
   	      sprintf(str, 
                "Read configuration RCfile: specify file name at bottom");
           if (cmd_type==3)
   	      sprintf(str, 
                "Reset the city file: specify file name at bottom");
           if (cmd_type==4)
   	      sprintf(str, 
                "Reset the vector map data file: specify file name at bottom");
	   if (cmd_type) drawWinString(scene, str, TOP);
	}

        if (cmd_mode==3) {
	   strcpy(str, MENU3);
	   drawWinString(scene, str, TOP);
	   for (i=1; i<=6; i++) {
	       j = XTextWidth(xfont[0], str, 13*i-1);
	       draw_separator(j);
	   }
	}
        if (cmd_mode==4) {
	   if (cmd_type==1)
              drawRectCmd(scene, param_descr, 17);
	   if (cmd_type==2)
              drawRectCmd(scene, font_descr, 6);
           if (cmd_type==3)
              drawRectCmd(scene, contin_descr, 5);
           if (cmd_type==4)
              drawRectCmd(scene, categ_descr, 4);
           if (cmd_type==5) {
	      strcpy(str, MENU4);             
	      drawWinString(scene, str, TOP);
	      for (i=1; i<=4; i++) {
	          j = XTextWidth(xfont[0], str, 18*i-1);
	          draw_separator(j);
	      }
	   }
	}
        if (cmd_mode==5) {
	  drawRectCmd(scene, feature_descr[cmd_type-1], 
                             feature_length[cmd_type-1]);
	}

        if (cmd_mode==1 || (cmd_mode==2 && cmd_type>=1) || 
            (cmd_mode==4 && cmd_type<=4) || cmd_mode==5)
              drawWinString(scene, cmd_string, BOTTOM);

	XFlush(dpy);
}

void image_init (struct imageLayout *s) {
	char str[256], compl[128];

	XFlush(dpy);
	usleep(20000);
	if (projection==SPHERICAL) 
           *compl = '\0';
	else
	   sprintf(compl, "aspect %.2f, ", aspect);

	sprintf(str, "%s %dx%d (proj.type %s, "
                    "zoom %.2f, %slat %.2f, lon %.2f, inc %.2f) ...",
                "Building map", s->width, s->height, 
		projtype[projection], 
                s->o.zoom, compl, s->o.xrot, s->o.yrot, s->o.zrot);
        drawWinString(s, str, TOP);

	if (s->im) XDestroyImage(s->im);
	s->im = XCreateImage(dpy, visual,
                DefaultDepth(dpy, scr), ZPixmap, 0, NULL, 
                s->width, s->height, color_pad, 0);
	s->im->data = (char *) malloc(s->im->bytes_per_line * s->im->height);
}

void printString (
#ifdef ZLIB
gzFile * fo,
#else
FILE * fo, 
#endif
char * s)
{
     if (fo)
#ifdef ZLIB
        gzprintf(fo, "%s", s);
#else
        fprintf(fo, "%s", s);
#endif
     else
        printf("%s", s);
}

void printPoint (
#ifdef ZLIB
gzFile * fo,
#else
FILE * fo, 
#endif
vector v)
{
     if (fo)
#ifdef ZLIB
        gzprintf(fo, "%7.3f %8.3f", -v.item[1]*coordconv, v.item[0]*coordconv);
#else
        fprintf(fo, "%7.3f %8.3f", -v.item[1]*coordconv, v.item[0]*coordconv);
#endif
     else
        printf("%7.3f %8.3f", -v.item[1]*coordconv, v.item[0]*coordconv);
}

void printImage (
#ifdef ZLIB
gzFile * fo,
#else
FILE * fo,
#endif
struct imageLayout * scene
)
{
   char str[120];
   int i, j, n, k, x, y, test, delta, step;
   Pixel p, bg;
   Pixel * pix;
   unsigned char *u, *v, *w, *code;
   short *inv;

   n = 0;
   for (i=0; i<5; i++)
     for (j=1; j<16; j++)
        if (scene->item[i][j].bool==ON) ++n;
   u = (unsigned char *) malloc(n*sizeof(unsigned char));
   v = (unsigned char *) malloc(n*sizeof(unsigned char));
   w = (unsigned char *) malloc(n*sizeof(unsigned char));
   pix = (Pixel *)malloc(n*sizeof(Pixel));

   n = 0;
   for (i=0; i<5; i++)
     for (j=1; j<16; j++)
        if (scene->item[i][j].bool==ON) {
	   u[n] = scene->item[i][j].r>>8;
	   v[n] = scene->item[i][j].g>>8;
	   w[n] = scene->item[i][j].b>>8;
	   if (reverse) {
	      u[n] = ~u[n];
	      v[n] = ~v[n];
	      w[n] = ~w[n];
	   }
	   pix[n] = scene->item[i][j].pix;
	   ++n;
	}
   bg = scene->item[CONF][C_BG_MAP].pix;

   k = 99;
   test = 1;
   while(test) {
   restart:
     k += 2;
     for (i=0; i<n-1; i++)
        for (j=i+1; j<n; j++) {
	   delta = pix[j]-pix[i];
	   if (delta!=0 && delta%k==0) goto restart;
        }
     test = 0;
   }

   inv = (short *)malloc(k*sizeof(short));
   code = (char *)malloc(n*sizeof(char));
   step = 0;
   for (i=0; i<n; i++) {
      j = ((unsigned long int)pix[i])%k;
      if (pix[i] == bg) {
	 code[i] = ' ';
	 step = 1;
      } else
         code[i] = 'A'+i-step;
      inv[j] = i;
   }

   if (img_output==1) {
      sprintf(str, "P6\n#Creator: xrmap (version %s)\n%d %d\n%d\n",
	   VERSION, scene->im->width, scene->im->height, 255);
      printString(fo, str);   
      for (y=0; y<scene->im->height; y++)
      for (x=0; x<scene->im->width; x++) {
	 p = XGetPixel(scene->im, x, y);
	 i = inv[((int)p)%k];
	 if (fo) {
#ifdef ZLIB
             gzputc(fo, u[i]);
             gzputc(fo, v[i]);
             gzputc(fo, w[i]);
#else
             fputc(u[i], fo);
             fputc(v[i], fo);
             fputc(w[i], fo);
#endif
	 } else {
	     fputc(u[i], stdout);
	     fputc(v[i], stdout);
	     fputc(w[i], stdout);
	 }
      }
   }

   if (img_output==2) {
      sprintf(str, "/* XPM */\n/* Creator: xrmap (version %s) */\n"
	          "static char * rmap_xpm[] = {\n"
	          "\"%d %d %d 1\",\n",
	   VERSION, scene->im->width, scene->im->height, n);
      printString(fo, str);   
      for (i=0; i<n; i++) if (code[i]==' ') {
	 sprintf(str, "\"%c\tc #%02X%02X%02X\",\n", code[i], u[i], v[i], w[i]);
         printString(fo, str);
      }
      for (i=0; i<n; i++) if (code[i]!=' ') {
	 sprintf(str, "\"%c\tc #%02X%02X%02X\",\n", code[i], u[i], v[i], w[i]);
         printString(fo, str);
      }
      for (y=0; y<scene->im->height; y++){
	 if (fo)
#ifdef ZLIB
            gzputc(fo, '\"');
#else
            fputc('\"', fo);
#endif         
	 else
	    fputc('\"', stdout);
         for (x=0; x<scene->im->width; x++) {
	    p = XGetPixel(scene->im, x, y);
	    i = inv[((int)p)%k];
	    if (fo)
#ifdef ZLIB
               gzputc(fo, code[i]);
#else
               fputc(code[i], fo);
#endif
	    else
	       fputc(code[i], stdout);
	 }
	 sprintf(str, "\"%c\n", (y==scene->im->height-1)? ';' : ',');
         printString(fo, str);
      }
   }

   free(u); free(v); free(w);
   free(pix); free(inv);
}

void createGCs(struct imageLayout *s) {
        XGCValues               gcv;
        int                     mask;
        mask = GCForeground | GCBackground | GCFont;
        gcv.background = s->item[CONF][C_BG_TEXT].pix;
	gcv.foreground = s->item[CONF][C_FG_TEXT].pix;
	gcv.font = xfont[0]->fid;
        gc = XCreateGC(dpy, Root, mask, &gcv);

	create_textpix(s->width);
        gcv.background = black;
        gcv.foreground = white;
        mask = GCForeground | GCBackground;
        textgc = XCreateGC(dpy, textpix, mask, &gcv);
}

void putPixel(XImage *im, int x, int y, Pixel pix)
{
        if (x>=0 && x<im->width && y>=0 && y<im->height)
	         XPutPixel(im, x, y, pix);
}

void drawPoint (struct imageLayout *s, vector a, Pixel color) {
	int x,y;

	x = round2int(a.item[X]);
	x += s->width/2;
	y = round2int(a.item[Y]);
	y += s->height/2;

        putPixel(s->im, x, y, color);
}

void drawString(XImage *im, int text_x, int text_y, char * text, 
		int which, Pixel pix)
{
        int i, j, l, textw;
	char u=0, test;
	char *ptr;
        XImage *xim;
      
	if (!text || !*text) return;
        XSetFont(dpy, textgc, xfont[which]->fid);
	ptr = text;

      redo:
	text = ptr;
	l = strlen(text);
	ptr = NULL;
	for (i=0; i<l-1; i++) if (text[i]=='\\' && text[i+1]=='n') {
	    text[i] = '\0';
	    ptr = text+i+2;
	    break;
	}
        XDrawImageString(dpy, textpix, textgc, 0, fontasc[which], 
                         text, strlen(text));
	textw = XTextWidth(xfont[which], text, strlen(text));
	if (textw>im->width) textw=im->width;
	
        xim = XGetImage(dpy, textpix, 0, 0, textw, fonth[which], 1, XYPixmap);
        if (!xim) return;
        test = (bigendian)? 128 : 1;
        for (j=0; j<fonth[which]; ++j) {
            for (i=0; i<textw; ++i) {
                if ((i&7) == 0) u = xim->data[j*xim->bytes_per_line+i/8];
                if (u&test) 
                   putPixel(im, text_x+i+1, text_y-fontasc[which]+j, pix);
                u = (bigendian)? u<<1 : u>>1;
            }
        }
	if (ptr) {
	   *(ptr-2) = '\\';
	   if (*ptr) {
	      text_y += fonth[which];
	      goto redo;
	   }
	}
        XDestroyImage(xim);
}

void drawSegment (struct imageLayout *s, vector a, vector b, Pixel color) {
  	int x, y, xp, yp, dx, dy,
            m, sum, min, max, addxmin, addxmax, addymin, addymax;
		
        /* lets try it without perspective */
        x = round2int(a.item[X]);
        xp = round2int(b.item[X]);
	dx = xp-x;
	min = s->width/2;
        x += min;

        y = round2int(a.item[Y]);
        yp = round2int(b.item[Y]);
        dy = yp-y;
	y += s->height/2;

        /* draw the line */
        if (dx>0)
	   addxmin = 1;
	else {
	   dx = -dx;
           addxmin = -1;
	}
	if (dx>min) return;

        if (dy>0)
	   addymin = 1;
	else {
	   dy = -dy;
           addymin = -1;
	}
	if (dx>dy) {
	  max = dx;
	  min = dy;
	  addxmax = addxmin;
	  addymax = 0;
	} else {
	  max = dy;
	  min = dx;
	  addxmax = 0;
	  addymax = addymin;
	}
        sum = (max/2) - max;
	for (m=0; m<max; m++) {
	   sum = sum+min;
	   if (sum>=max/2) {
              sum -= max;
              x += addxmin;
              y += addymin;
           } else {
              x += addxmax;
              y += addymax;
           }
	   putPixel(s->im, x, y, color);
	}
}   

Window makeWindow (struct imageLayout *s) {
        Window w;
        black = BlackPixel(dpy, scr);
        white = WhitePixel(dpy, scr);

	w = XCreateSimpleWindow(dpy, Root,
                      10, 10,
                      s->width, s->height, 0,
                      white, black);
        XSetWindowColormap(dpy, w, cmap);

        XSelectInput(dpy, w, ButtonPressMask | ButtonReleaseMask |
                             PointerMotionMask | ExposureMask | 
                             KeyReleaseMask | StructureNotifyMask);

        wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
        wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
        XSetWMProtocols(dpy, w, &wm_delete_window, 1);

        show_title(w);
        return w;
}

Bool
evpred(d, e, a)
Display *              d;
XEvent *               e;
XPointer               a;
{
        return (True);
}

static vector project (vector a, matrix *b) {
	vector v;
	int i;
	
	if (projection==SPHERICAL) {
           v.item[0] = a.item[0]*b->item[0] + a.item[1]*b->item[3] + a.item[2]*b->item[6];
	   v.item[1] = a.item[0]*b->item[1] + a.item[1]*b->item[4] + a.item[2]*b->item[7];
	   v.item[2] = a.item[0]*b->item[2] + a.item[1]*b->item[5] + a.item[2]*b->item[8];
	} else {
           v.item[0] = a.item[0] - b->item[0];
	   if (v.item[0]>earth_radius) {
	      i = (int)(v.item[0]/earth_radius);
	      i = (i+1)&(-2);
              v.item[0] -= ((double)i)*earth_radius;
	   } else
	   if (v.item[0]<-earth_radius) {
	      i = (int)(-v.item[0]/earth_radius);
	      i = (i+1)&(-2);
              v.item[0] += ((double)i)*earth_radius;
	   }
           v.item[0] *= b->item[2];
           v.item[1] = (a.item[1] - b->item[1]) * b->item[2];
           v.item[2] = a.item[2] * b->item[2];	  
	}
	return v;
}

void operate(double *y)
{
        if (projection==MERCATOR) {
	   if (*y>0.0) {
              if (*y>0.495) 
                 *y = 1.55;
	      else
                 *y = ONE_OVER_PI*log(tan(PI_OVER_2*(*y+0.5)));
	   } else {
              if (*y<-0.495) 
                 *y = -1.55;
	      else
                 *y = -ONE_OVER_PI*log(tan(PI_OVER_2*(-*y+0.5)));
	   }
	}
	if (projection==MILLER)
           *y = FIVE_OVER_4PI*log(tan(PI_OVER_2*(0.8*(*y)+0.5)));
}

void inverse_operate(double *y)
{
        if (projection==MERCATOR)
	   *y = atan(exp(*y/ONE_OVER_PI))/PI_OVER_2-0.5;
	if (projection==MILLER)
	   *y = 1.25*(atan(exp(*y/FIVE_OVER_4PI))/PI_OVER_2-0.5);
}

static vector spher2cart (spherePt s) {
        vector v;
	double latrad, lonrad, y;

	if (projection==SPHERICAL) {
	   latrad = s.item[LATITUDE] * CONV;
	   lonrad = s.item[LONGITUDE] * CONV;
	   y = s.item[RADIUS] * cos(latrad);

           v.item[X] = y * sin(lonrad);
           v.item[Y] = s.item[RADIUS] * sin(latrad);
           v.item[Z] = y * cos(lonrad);
           return v;
	} else {
	   v.item[X] = s.item[RADIUS] * s.item[LONGITUDE] / 180.0;
	   y = -s.item[LATITUDE] / 180.0;
	   operate(&y);
	   v.item[Y] = s.item[RADIUS] * y * aspect ;
           v.item[Z] = 1.0;
	   return v;
	}
}

static int inverse_coord(struct imageLayout * s, 
                     int i, int j, double *x, double *y)
{
        double fact;
        vector a, v;
	matrix *b;
	if (i<0 || j<0 || i>=s->width || j>=s->height) return 0;
	fact = earth_radius*s->o.zoom;
        a.item[X] = ((double)(i-s->width/2))/fact;
	if (projection==SPHERICAL) {
           a.item[Y] = ((double)(j-s->height/2))/fact;
	   a.item[Z] = 1-(a.item[X])*(a.item[X])-(a.item[Y])*(a.item[Y]);
	   if (a.item[Z]<0) return 0;
	   a.item[Z] = sqrt(a.item[Z]);
	   b = &s->rotation;
           v.item[0] = a.item[0]*b->item[0] + a.item[1]*b->item[1] + a.item[2]*b->item[2];
	   v.item[1] = a.item[0]*b->item[3] + a.item[1]*b->item[4] + a.item[2]*b->item[5];
	   v.item[2] = a.item[0]*b->item[6] + a.item[1]*b->item[7] + a.item[2]*b->item[8];

	   v.item[0] /= s->o.zoom;
	   v.item[1] /= s->o.zoom;
	   v.item[2] /= s->o.zoom;
	   *x = atan2(v.item[1], sqrt(v.item[0]*v.item[0]+v.item[2]*v.item[2]))
                     / CONV;
	   *y = atan2(v.item[0] , v.item[2]) / CONV;
	   return 1;
	} else {
           *y = a.item[X] * 180.0 + s->o.yrot;
	   while (*y > 180.0) *y -= 360.0;
	   while (*y <= -180.0) *y += 360.0;
           a.item[Y] = ((double)(j-s->height/2))/(fact*aspect)
                       + s->rotation.item[1]/(earth_radius * aspect);
	   inverse_operate(&a.item[Y]);
	   *x = -a.item[Y] * 180.0;
	   if (fabs(*x)>=90.0) return 0;
	   return 1;
	}
}

static int segment_in_scene (struct imageLayout *s, struct segment_index *i) {
	spherePt max, min;
	vector pt1, pt2;
	int bit_number[9] = {0,1,2,0,3,0,0,0,4};

	/* first verify if we want this printed */
	if (s->item[bit_number[i->category]][i->type].bool != ON)
		return 0;

	max.item[LONGITUDE] = i->maxlong/3600.0;
	max.item[LATITUDE] = i->maxlat/3600.0;
	max.item[RADIUS] = earth_radius;

	min.item[LONGITUDE] = i->minlong/3600.0;
	min.item[LATITUDE] = i->minlat/3600.0;
	min.item[RADIUS] = earth_radius;

	pt1 = project(spher2cart(max), &s->rotation);
	pt2 = project(spher2cart(min), &s->rotation);

	/* check to see if it is z-clipped */
	if (s->transparent == 0 && (pt1.item[Z] < 0 || pt2.item[Z] < 0)) {
		return 0; 
	}
		
	/* Yeah, this is ugly, but my head hurts right now. I've got the flu
	   I'll try to figure it out later */

	if (    (pt1.item[X] > s->width/2       && pt2.item[X] > s->width/2)          ||
		(pt1.item[Y] > s->width/2       && pt2.item[Y] > s->width/2)          || 
		(pt1.item[X] < (-1.*s->width/2) && pt2.item[X] < (-1.*s->width/2) )   ||
		(pt1.item[Y] < (-1.*s->width/2) && pt2.item[Y] < (-1.*s->width/2) ) ) {
		return 0;
	}
	
	/* See if the whole segment can fit into one point, 
           if so, then just draw the point! */
	if ( (int)pt1.item[X] == (int)pt2.item[X] && (int)pt1.item[Y] == (int)pt2.item[Y] ) {
		return POINT_ONLY;
	}

	return 1;
}


void drawCircles(struct imageLayout *s) {
        int i;
	double x, y; 
	spherePt spt1, spt2;
	vector pt1, pt2;	
	double step=2.0; /* small step */

	spt1.item[RADIUS] = earth_radius;
	spt2.item[RADIUS] = earth_radius;

	if (s->item[CONF][C_LON_GRID].bool == ON) 
	for (i=0; i<=1; i++)
	for (x=i*s->o.spacing_lon; x<180.001-i*(s->o.spacing_lon/5); 
             x+=s->o.spacing_lon) {
		for (y=-90.0; y<=90.1-step; y+=step) {
			spt1.item[LONGITUDE] = (i==0)? x : -x;
			spt1.item[LATITUDE]  = y;
			spt2.item[LONGITUDE] = (i==0)? x : -x;
			spt2.item[LATITUDE]  = y+step;
			pt1 = project(spher2cart(spt1), &s->rotation);
			pt2 = project(spher2cart(spt2), &s->rotation);
			if (pt1.item[Z] > 0) 
				drawSegment(s, pt1, pt2, s->item[CONF][C_LAT_GRID].pix);
		}
	}

	if (s->item[CONF][C_LAT_GRID].bool == ON) 
	for (i=0; i<=1; i++)
	for (x=i*s->o.spacing_lat; x<90.0-s->o.spacing_lat/5; 
             x+=s->o.spacing_lat) {
		for (y=-180.0; y<180.1-step; y+=step) {
			spt1.item[LONGITUDE] = y;
			spt1.item[LATITUDE]  = (i==0)? x : -x;
			spt2.item[LONGITUDE] = y+step;
			spt2.item[LATITUDE]  = (i==0)? x : -x;
			pt1 = project(spher2cart(spt1), &s->rotation);
			pt2 = project(spher2cart(spt2), &s->rotation);
			if (pt1.item[Z] > 0) 
				drawSegment(s, pt1, pt2, s->item[CONF][C_LON_GRID].pix);
		}
	}

	if (s->item[CONF][C_EARTH_CONTOUR].bool == ON &&
	    projection==SPHERICAL) {
	   y = earth_radius * s->o.zoom;
	   pt2.item[X] = y;
	   pt2.item[Y] = 0;
	   for (x=step; x<360.1-step; x+=step) {
	       pt1 = pt2;
	       pt2.item[X] = y * cos(x*CONV);
	       pt2.item[Y] = y * sin(x*CONV);
	       drawSegment(s, pt1, pt2, s->item[CONF][C_EARTH_CONTOUR].pix);
	   }
	}
}

int size_of_city(int n)
{
    int i;    
    for (i=0; i<5; i++)
        if (n>popul_step[i]) return i;
    return 0;
}

void drawLocation (struct imageLayout *s, struct location * loc) {
	vector pt;
        spherePt spt;
	int x, y, i, j, d, u, ind=0, size;
	int text_x = 0, text_y = 0;
	Pixel pixel = black;
        unsigned short * bits;
	char name[128];
	char *ptr = NULL, *seatype = NULL;

	i = loc->tz;
	if (i<0) return;
	if (loc->obj=='c')
	   i = 0x00000001<<timezones[i]->continent;
	if (loc->obj=='l')
	   i = 0x00000001<<(i-1);
	if (!(s->o.desired_continents & i)) return;

	if (loc->obj=='c') {
	  if (s->item[CONF][C_CITY_CIRCLE].bool!=ON && 
              s->item[CONF][C_CITY_TEXT].bool!=ON)
              return;
	}
	if (loc->obj=='l') {
	  seatype = index("BKOSTV", loc->status);
	  if (s->item[CONF][C_LOC_LAND].bool!=ON && !seatype) return;
	  if (s->item[CONF][C_LOC_WATER].bool!=ON && seatype) return;
	}

	spt.item[RADIUS] = earth_radius;
	spt.item[LONGITUDE] = loc->lon;
	spt.item[LATITUDE] = loc->lat;

	/* determine the point, post transformation */
	pt = project(spher2cart(spt), &s->rotation);

	if (pt.item[Z] < 0) 
		return;

	x = pt.item[X] + s->width/2;
	y = pt.item[Y] + s->height/2;

	/* verify that the point is on the map, return if not */
	if (x > s->width || x < 0 || y > s->height || y < 0)
		return;
	
	/* draw a circle around the point in question if it's a city */
	if (loc->obj=='c') {
	   if (!index(city_filter, loc->status)) return;
	   size = size_of_city(loc->population);
	   ind = spot_size[size]-1;
	   if (ind<0) return;
	   if (zoom_value[size] > 
               (int)(2*s->o.zoom*earth_radius)) return;
	   loc->drawn = '+';
	   bits = spot_bits[ind];
	   d = bits[0]/2;
	   if (s->item[CONF][C_CITY_CIRCLE].bool == ON) {
	      pixel = s->item[CONF][C_CITY_CIRCLE].pix;
	      for (j=0; j<bits[0]; j++) {
                  if (y-d+j >= s->height) break;
                  u = bits[j+1];
                  for (i=0; i<bits[0]; i++) {
                  if (u&1) putPixel(s->im, x-d+i, y-d+j, pixel);
                  u = u>>1;
	          }
	      }
	   }
	   /* determine the text location relative to the circle */
	   d += 1;
	   ind = size;
	   if (s->item[CONF][C_CITY_TEXT].bool != ON) return;
	   if ((ptr=index(loc->name,'('))) {
	      strcpy(name, ptr+1);
	      ptr = index(name,')');
	      if (ptr) *ptr = '\0';
	   } else
	      strcpy(name, loc->name);
	   ptr = name;
	   text_x = (place_name && x > s->width/2) ? 
	     (x - XTextWidth(xfont[ind], name, strlen(name)) - d) :
	     (x + d);
	   text_y = (place_name && y < s->height/2) ? (y+fonth[ind]) : (y - d);
	   pixel = s->item[CONF][C_CITY_TEXT].pix;
	} 

	if (loc->obj=='l') {
	   ind = loc->population;
	   if (!ind) return;
           if (!index(loc_filter, loc->status)) return;
	   ptr = loc->name;
	   text_x = x;
	   text_y = y;
	   if (loc->rank) {
              char *ptr1, *ptr2;
	      int i1;
	      i = 0;
	      if (loc->rank<=3)
		 j = fontasc[ind]-fonth[ind]-1;
	      else
	      if (loc->rank>=5)
		 j = fontasc[ind]+1;
	      else
		 j = 0;
	      ptr1 = ptr;
	      while(ptr1) {
		 ptr2 = strstr(ptr1, "\\n");
		 if (!ptr2) ptr2 = ptr1+strlen(ptr1);
                 i1 = XTextWidth(xfont[ind], ptr1, (int)(ptr2-ptr1));
		 if (i1>i) i = i1;
		 if (*ptr2=='\\') {
                    ptr1 = ptr2+2;
		    if (loc->rank<=3) j -= fonth[ind];
		 } else
		    ptr1 = NULL;
	      }
	      if (loc->rank>=3 && loc->rank<=5) ++i;
	      else
	      if (loc->rank==2 || loc->rank==6) i = i/2;
	      else
		 i = -1;
	      text_x -= i; 
              text_y += j;
	   }
	   if (seatype)
	      pixel = s->item[CONF][C_LOC_WATER].pix;
           else
	      pixel = s->item[CONF][C_LOC_LAND].pix;
	}

	/* add text */
	if (!textpix) create_textpix(s->width);
	drawString(s->im, text_x, text_y, ptr, ind, pixel);
}

/* This function will generate the proper transformation 
 * matrix (rotation and scaling) and copy the values 
 * to the imageLayout struct
 */

void computeMatrix(struct imageLayout *s) {
     double y, cosx, cosy, cosz, sinx, siny, sinz;

     if (projection==SPHERICAL) {
        /* generate the proper rotation & scale matrix */
        cosx = cos(s->o.xrot * CONV);
        cosy = cos(s->o.yrot * CONV);
        cosz = cos(s->o.zrot * CONV);

        sinx = sin(s->o.xrot * CONV);
        siny = sin(s->o.yrot * CONV);
        sinz = sin(s->o.zrot * CONV);

	s->rotation.item[0] = s->o.zoom*(cosy*cosz + sinx*siny*sinz);
	s->rotation.item[1] = s->o.zoom*(-cosy*sinz + sinx*siny*cosz);
	s->rotation.item[2] = s->o.zoom*(cosx*siny);

	s->rotation.item[3] = s->o.zoom*(-cosx*sinz);
	s->rotation.item[4] = s->o.zoom*(-cosx*cosz);
	s->rotation.item[5] = s->o.zoom*(sinx);

	s->rotation.item[6] = s->o.zoom*(-siny*cosz + sinx*cosy*sinz);
	s->rotation.item[7] = s->o.zoom*(siny*sinz + sinx*cosy*cosz);
	s->rotation.item[8] = s->o.zoom*(cosx*cosy);

     } else {
        s->rotation.item[0] = s->o.yrot * earth_radius / 180.0;
        y = s->o.xrot / 180.0;
	operate(&y);
        s->rotation.item[1] = - earth_radius * y * aspect;
        s->rotation.item[2] = s->o.zoom;
     }
}
	
void openOutfile(struct imageLayout * scene,
#ifdef ZLIB
gzFile **fo
#else
FILE **fo
#endif
)
{
        int i;
	struct stat buf;
	char str[256], strgz[256];

	if (!scene->o.outfile || !*scene->o.outfile) {
	   *fo = NULL;
	   overwrite = -1;
	   return;
	}

        i = strlen(scene->o.outfile)-3;
        if (strcmp(scene->o.outfile+i, ".gz")) {
	   sprintf(str,"%s", scene->o.outfile);
	   sprintf(strgz,"%s.gz", scene->o.outfile);
        } else {
	   strcpy(strgz, scene->o.outfile);
           strcpy(str, strgz);
	   str[i-3] = '\0';
        }
        if (overwrite==0) {
           stat(str, &buf);
	   if (S_ISREG(buf.st_mode)) overwrite = 1;
           stat(strgz, &buf);
           if (S_ISREG(buf.st_mode)) overwrite = 1;
	   if (overwrite==1) return;
	}
 
        #ifdef ZLIB
	*fo = gzopen(strgz, "wb6");
	strcpy(str, strgz);
	#else
	*fo = fopen(str, "w");
	#endif
	if (!*fo)
	   fprintf(stderr, "Opening %s for writing failed !!\n", str);
	overwrite = -1;
}

void closeOutfile(
#ifdef ZLIB
gzFile *fo
#else
FILE *fo
#endif
)
{
	if (fo) {
	   #ifdef ZLIB
	   gzclose(fo);
	   #else
	   fclose(fo);
	   #endif
	}
}

void generate (struct imageLayout *scene) {
        /* misc variables */
	spherePt s1,s2;
	vector pt1, pt2;
	int i,j,ret;
	/* filehandle */
#ifdef ZLIB
	gzFile * fh; 
        gzFile * fo = NULL;
#else
	int fh;
        FILE * fo = NULL;
#endif
	/* filenames */
	struct header header_data;
	struct segment_index segment_index_data;
	struct segment_header segment_header_data;
	struct stroke stroke_data;
	/* simple mapping array */
	int bit_number[9] = {0,1,2,0,3,0,0,0,4};
	char str[256];
        Pixel bg;

	cmd_string = realloc(cmd_string, 2);
	*cmd_string = '\0';
	caret_pos = 0;
        change_made = 0;
	color_change = 0;
	
	/* open outfile if necessary */
	if (vectordata || img_output)
	   openOutfile(scene, &fo);

	/* initialize the image stuff */
	image_init(scene);

	/* compute the rotation matrix */
	computeMatrix(scene);

	/* set the background color */
        bg = scene->item[CONF][C_BG_MAP].pix;
	for (i=0; i<scene->width; i++)
	    for (j=0; j<scene->height; j++)
	        XPutPixel(scene->im, i, j, bg);

        drawCircles(scene);
	
#ifdef DEBUG
	fprintf(stderr,"-- GENERATE --\n");
	fprintf(stderr,"o Options\n");
	fprintf(stderr,"  - slatitude   [%d]\n", scene->o.spacing_lat);
	fprintf(stderr,"  - slongitude  [%d]\n", scene->o.spacing_lon);
	fprintf(stderr,"  - zoom        [%g]\n", scene->o.zoom);
	fprintf(stderr,"  - aspect      [%g]\n", aspect);
	fprintf(stderr,"  - latitude    [%g]\n", scene->o.xrot);
	fprintf(stderr,"  - longitude   [%g]\n", scene->o.yrot);
	fprintf(stderr,"  - inclination [%g]\n", scene->o.zrot);
	fprintf(stderr,"  - projection  [%d]\n", projection);
	fprintf(stderr,"  - mapfile     [%s]\n", scene->o.mapfile);
        fprintf(stderr,"  - rcfile      [%s]\n", scene->o.rcfile);
	fprintf(stderr,"  - datafile     [%s]\n", scene->o.datafile);
	fprintf(stderr,"  - des_cat     [%X]\n", scene->o.desired_categories);
	fprintf(stderr,"  - des_cont    [%X]\n", scene->o.desired_continents);
	fprintf(stderr,"o imageLayout (Scene)\n");
	fprintf(stderr,"  - height      [%d]\n", scene->height);
	fprintf(stderr,"  - width       [%d]\n", scene->width);
	fprintf(stderr,"  - transp.     [%d]\n", scene->transparent);
	fprintf(stderr,"  o Colormap\n");
	for (i=0; i<5; i++) 
		for (j=0; j<16; j++) 
			if (scene->item[i][j].bool != INACTIVE)
				fprintf(stderr,"    - (%2d,%2d) [%s]  %s\n",i ,j , (scene->item[i][j].bool)? "on":"off", scene->item[i][j].name);
	fprintf(stderr,"  o Rotation Matrix\n");
	for (i=0; i<3; i++) 
		fprintf(stderr,"     [%e][%e][%e]\n", scene->rotation.item[i][0],  
			scene->rotation.item[i][1], scene->rotation.item[i][2]);
	fprintf(stderr,"-- -------- --\n");
#endif DEBUG

	/* Open the datafile and find graph the appropriate sectors */
#ifdef ZLIB
	fh = gzopen(scene->o.mapfile, "r");
	if (fh == NULL) {
		fprintf(stderr,"Couldn't open input vector file `%s'\n", scene->o.mapfile);
		exit (-1);
	}
#else
        fh = open(scene->o.mapfile, O_RDONLY, 0);
        if (fh == -1) {
		fprintf(stderr,"Couldn't open input vector file `%s'\n", scene->o.mapfile);
		exit (-1);
	}
#endif
	/* read the header */
#ifdef ZLIB
	gzread(fh, &header_data, sizeof(header_data));
#else
	read(fh, &header_data, sizeof(header_data));
#endif
#ifdef WORDS_BIGENDIAN
	bitflip( &(header_data.magic) );	
	bitflip( &(header_data.segment_index_address) );	
	bitflip( &(header_data.segment_index_count) );	
#endif /* WORDS_BIGENDIAN */
	if (header_data.magic != HEADER_MAGIC) {
		fprintf(stderr, "The magic was not valid for the input file.\n");
		fprintf(stderr, "This could be an endian problem, or you\n"); 
		fprintf(stderr, "could be pointing to the wrong file.\n");
		fprintf(stderr, "(%s's magic = %X and it's supposed to be %X)\n", 
			scene->o.mapfile, header_data.magic, HEADER_MAGIC);
		exit(-2);
	}
	/* iterate through all the segment_indexes to see 
           which ones are interesting */
	for (i=0; i<header_data.segment_index_count; i++) {
#ifdef ZLIB
		gzseek(fh, sizeof(header_data) + sizeof(segment_index_data)*i, SEEK_SET);
		gzread(fh, &segment_index_data, sizeof(segment_index_data));
#else
		lseek(fh, sizeof(header_data) + sizeof(segment_index_data)*i, SEEK_SET);
		read(fh, &segment_index_data, sizeof(segment_index_data));
#endif

#ifdef WORDS_BIGENDIAN
		bitflip( &(segment_index_data.maxlat) );
		bitflip( &(segment_index_data.minlat) );
		bitflip( &(segment_index_data.maxlong) );
		bitflip( &(segment_index_data.minlong) );
		bitflip( &(segment_index_data.segment_address) );
		bitflip( &(segment_index_data.continent) );
		bitflip( &(segment_index_data.category) );
		bitflip( &(segment_index_data.type) );
#endif /* WORDS_BIGENDIAN */

		if ( (segment_index_data.continent & scene->o.desired_continents)
		 && (segment_index_data.category & scene->o.desired_categories) ) { 
		        /* it has interesting information, 
                           let's look to see if the
			   data is in our viewport
			*/
			ret = segment_in_scene(scene, &segment_index_data);
			if (ret == YES) {
				/* iterate through the data and pass 
                                   it to draw */
#ifdef ZLIB
				gzseek(fh, segment_index_data.segment_address, SEEK_SET);
				gzread(fh, &segment_header_data, sizeof(segment_header_data));
#else
				lseek(fh, segment_index_data.segment_address, SEEK_SET);
				read(fh, &segment_header_data, sizeof(segment_header_data));
#endif
#ifdef WORDS_BIGENDIAN
				bitflip( &(segment_header_data.orgx) );
				bitflip( &(segment_header_data.orgy) );
				bitflip( &(segment_header_data.nstrokes) );
#endif /* WORDS_BIGENDIAN */
				s1.item[LONGITUDE] = (segment_header_data.orgx/3600.0);
				s1.item[LATITUDE]  = (segment_header_data.orgy/3600.0);
				s1.item[RADIUS] = earth_radius;
				s2.item[RADIUS] = earth_radius;

				for (j=0; j<segment_header_data.nstrokes; j++) {
				        /* iterate and draw all the 
                                           appropriate points */
#ifdef ZLIB
					gzread(fh, &stroke_data, sizeof(stroke_data));
#else
					read(fh, &stroke_data, sizeof(stroke_data));
#endif

#ifdef WORDS_BIGENDIAN
					bitflip( &(stroke_data.dx) );
					bitflip( &(stroke_data.dy) );
#endif /* WORDS_BIGENDIAN */
					s2.item[LONGITUDE] = s1.item[LONGITUDE]+(stroke_data.dx/3600.0);
					s2.item[LATITUDE]  = s1.item[LATITUDE] +(stroke_data.dy/3600.0);

					pt1 = project(spher2cart(s1), &scene->rotation);
					pt2 = project(spher2cart(s2), &scene->rotation);
	
					drawSegment(scene, pt1, pt2, 
						scene->item[bit_number[segment_index_data.category]][segment_index_data.type].pix);
					if (vectordata>0) {
                                	   if (j==0) {
					      sprintf(str, "#%d %d\n", 
                                                numcurves,
                                                10*segment_index_data.type+
					        +segment_index_data.category);
					      printString(fo, str);
					      ++numcurves;
					      printPoint(fo, pt1);
					      printedpt = 1;
					      numpt = 0;
					      
					   }
					   ++numpt;
					   if (numpt%vectordata==0 || j==
                                              segment_header_data.nstrokes-1) {
					      if (printedpt%4==0)
						 sprintf(str, "\n");
					      else
						 sprintf(str, "  ");
					      printString(fo, str);
					      printPoint(fo, pt2);
					      ++printedpt;
					   }
					}
					s1.item[LONGITUDE] = s2.item[LONGITUDE];
					s1.item[LATITUDE]  = s2.item[LATITUDE];
				}
				if (vectordata>0)
                                   printString(fo, "\n;\n\n");
			} else if (ret == POINT_ONLY) {
				/* the whole segment amounts to a point 
                                   based on the perspective of the scene, 
                                   so just draw a point	*/
				s1.item[LONGITUDE] = segment_index_data.maxlong/3600.;
				s1.item[LATITUDE] = segment_index_data.maxlat/3600.;
				s1.item[RADIUS] = earth_radius;
				pt1 = project(spher2cart(s1), &scene->rotation);
				drawPoint(scene, pt1, 
					scene->item[bit_number[segment_index_data.category]][segment_index_data.type].pix);
				if (vectordata<0) {
				   printPoint(fo, pt1);
				   printString(fo, "\n");
				}
			}
		}
	}

	/* iterate through locations and show spot/texts */
	for (i=0; i<numloc; i++) {
	      locations[i]->drawn = '\0';
	      drawLocation(scene, locations[i]);
	}

        putPicture(scene);

	if (img_output) printImage(fo, scene);
        closeOutfile(fo);
	XSync(dpy, True);
}

void redisplay(struct imageLayout * scene, int force)
{
	if (force || cmd_mode<=3 || (cmd_mode==4 && cmd_type==5)) 
           putPicture(scene);
	if (cmd_mode>0) drawWinCmd(scene);
}

void do_escape(struct imageLayout * scene)
{
	line_clicked = -1;
	if (cmd_mode<=3)
	   cmd_mode = 0;
	else {
	   cmd_mode -= 1;
	   *cmd_string = '\0';
	   caret_pos = 0;
	   if (cmd_mode == 4) cmd_type = 5;
	}
        show_title(win);
	if (color_change) get_all_colors(scene);
	if (cmd_mode==0 && change_made) {
	   XResizeWindow(dpy, win, scene->width, scene->height);
	   create_textpix(scene->width);
	   XSync(dpy, True);
           set_radius(scene);
	   generate(scene);
	}
        redisplay(scene, 1);
}

void draw_flag(struct imageLayout * scene, XImage *image, int u, int v, int p)
{
     int x, y;

     if (u>scene->width/2) {
        x = 3+ (p/2) * (image->width+5);
     } else {
        x = scene->width - 3 - image->width - ((p==1)?(image->width+5) : 0);
     }

     if (v>scene->height/2) 
        y = scene->height - image->height - 3;
     else
        y = 3;

     XSetWindowBackground(dpy, win, scene->item[CONF][C_BG_TEXT].pix);
     XClearArea(dpy, win, x-3, y-3, image->width+6, image->height+6,False);
     XPutImage(dpy, win, gc, image, 0, 0, x, y, 
               image->width, image->height);
}

void show_point(struct imageLayout * scene, int u, int v)
{
        XpmAttributes attrib;
        XImage *image = NULL;
	char str[256], strp[256], strs[256],
	     path1[256], path2[256],
             name[256], compl1[40], compl2[40], 
             alt[20], pop[20], rank1[20], rank2[20], adm_status[30];
	char *ptr, *ptrp;
        double x, y, eps;
	int i=0, j, k, bool, found, mode, popul, max;
	int up, vp, uf=0, vf=0;

	mode = cmd_mode;
	cmd_mode = 0;
	if (cmd_mode==0 && but_pressed==3) {
	   strcpy(compl1, "Recenter map to point (");
	   strcpy(compl2, ") ?  [Cancel: move out of window]");
	} else
	   *compl1 = *compl2 = '\0';

        if (inverse_coord(scene, u, v, &x, &y)) {
	   found = -1;
	   eps = 1.5/scene->o.zoom;
	   if (but_pressed==1) {
	     popul = 0;
	     max = 10;
	     for (i=0; i<numloc; i++) {
	       if (locations[i]->obj=='c' && locations[i]->drawn=='+' &&
	          fabs(x-locations[i]->lat)+fabs(y-locations[i]->lon)<eps) {
		 vector pt;
                 spherePt spt;
         	 spt.item[RADIUS] = earth_radius;
	         spt.item[LONGITUDE] = locations[i]->lon;
	         spt.item[LATITUDE] = locations[i]->lat;
		 pt = project(spher2cart(spt), &scene->rotation);
		 up = pt.item[X] + scene->width/2;
                 vp = pt.item[Y] + scene->height/2;
		 j = abs(u-up);
		 k = abs(v-vp);
                 if (j<3 && k<3 && j+k<=max) {
		    if (j+k<max) popul = 0;
		    max = j+k;
		    if (locations[i]->population>popul) {
		       found = i;
		       uf = up;
		       vf = vp;
                       popul = locations[i]->population;
		    }
		 }
	       }
	     }
	   }
           if (city_shown && but_pressed==1) {
	      if (found==-1) return;
	      if (found==city_found) return;
           }

	   bool = 0;
	   if (v>scene->height/2) {
	      if (point_shown && line_pos==BOTTOM) bool = 1;
	      j = TOP;
	   } else {
	      if (point_shown && line_pos==TOP) bool = 1;
	      j = BOTTOM;
	   }
	   if (bool || city_shown) putPicture(scene);
	   line_pos = j;
	   if (found>=0) {
	      city_shown = 1;
	      point_shown = 1;
	      city_found = found;
	      XSetForeground(dpy, gc, scene->item[CONF][C_CITY_MARKED].pix);
	      XDrawLine(dpy, win, gc, uf-6, vf, uf+6, vf);
	      XDrawLine(dpy, win, gc, uf, vf-6, uf, vf+6);
              XSetForeground(dpy, gc, scene->item[CONF][C_FG_TEXT].pix);
	      if (locations[found]->alt==-30000) 
		 strcpy(alt, "?");
	      else
                 sprintf(alt, "%dm", locations[found]->alt);
	      if (locations[found]->status=='C') 
                 strcpy(adm_status, "   (capital of state)");
	      else
	      if (locations[found]->status=='R') 
                 strcpy(adm_status, "   (capital of region)");
	      else
		 *adm_status = '\0';
	      k = 0;
	      for (j=0; j<=strlen(locations[found]->name); j++) {
		  if (locations[found]->name[j]=='(') {
		     name[j] = ' ';
		     k = 1;
		  }
		  name[j+k] = locations[found]->name[j];
	      }
	      sprintf(str, "%s, %s   lat = %.2f  lon = %.3f   alt = %s%s",
		    name, locations[found]->country,
		    locations[found]->lat, locations[found]->lon, 
                    alt, adm_status);
	      if (locations[found]->population) 
                 sprintf(pop, "%d", locations[found]->population);
	      else
		 strcpy(pop, "?");
	      if (locations[found]->rank) 
                 sprintf(rank1, "%d", locations[found]->rank);
	      else
		 strcpy(rank1, "-");
	      if (locations[found]->worldrank) 
                 sprintf(rank2, "%d", locations[found]->worldrank);
	      else
		 strcpy(rank2, "-");

	      sprintf(strp, "Population = %s  (rank: state/%s world/%s)  "
                            "administrative region : %s",
		 pop, rank1, rank2,
                 (locations[found]->region[0])?locations[found]->region:"-");
	      j = locations[found]->tz;
	      sprintf(strs, "Area/timezone: %s ; GMT%s",
		 timezones[j]->name, timezones[j]->tzvalue);
	   } else
	      sprintf(str, "%slat = %.3f  lon = %.3f%s", 
                      compl1, x, y, compl2);
	   if (found==-1 && !city_shown) {
	      point_shown = 1;
	      drawWinString(scene, str, line_pos);
	   } else {
	     ptr = timezones[j]->flags;
	     if ((ptrp = index(ptr, ','))) {	
                *ptrp = '\0';
		sprintf(path1, "%s/%s.xpm.gz", FLAGPATH, ptr);
		sprintf(path2, "%s/%s.xpm.gz", FLAGPATH, ptrp+1);
		*ptrp = ',';
	     } else {
		sprintf(path1, "%s/%s.xpm.gz", FLAGPATH, ptr);
                *path2 = '\0';
	     }
	     attrib.colormap = cmap;
             attrib.valuemask = XpmColormap | XpmReturnPixels;
             if (XpmReadFileToImage(dpy, path1, &image, NULL, &attrib)) {
	        fprintf(stderr, 
                  "Cannot read XPM file %s !!\n", path1);
	     } else {
	        draw_flag(scene, image, u, v, (*path2)?1:0);
	        XDestroyImage(image);
	     }
	     if (*path2) {
	        attrib.colormap = cmap;
                attrib.valuemask = XpmColormap | XpmReturnPixels;
                if (XpmReadFileToImage(dpy, path2, &image, NULL, &attrib)) {
	           fprintf(stderr, 
                     "Cannot read XPM file %s !!\n", path2);
	        } else {
	           draw_flag(scene, image, u, v, 2);
	           XDestroyImage(image);
		}
	     }
	     if (line_pos==TOP) {
	        drawWinString(scene, str, TOP);
	        drawWinString(scene, strp, PRETOP);
	        drawWinString(scene, strs, PREPRETOP);
	     } else {
	        drawWinString(scene, str, PREPREBOTTOM);
	        drawWinString(scene, strp, PREBOTTOM);
	        drawWinString(scene, strs, BOTTOM);
	     }
	   }
	}
	cmd_mode = mode;
}

void do_caret(struct imageLayout * scene, int u)
{
        int l;
        l = strlen(cmd_string);
        caret_pos = 0;
        while (caret_pos<l &&
           XTextWidth(xfont[0], cmd_string, caret_pos+1)<u-4) ++caret_pos;
	drawWinString(scene, cmd_string, BOTTOM);
        XFlush(dpy);
}

void filemenu(struct imageLayout * scene)
{
	overwrite = 0;
	cmd_mode = 2;
	cmd_type = 0;
	show_title(win);
	cmd_string = (char *)realloc(cmd_string, 1);
	*cmd_string = '\0';
	caret_pos = 0;
	redisplay(scene, 0);
}

void commands(struct imageLayout * scene)
{
    cmd_mode = 1;
    show_title(win);
    redisplay(scene, 0);
}

int main (int argc, char **argv) {
        /* Global geometric structure */
        struct imageLayout *scene; 
	/* misc variables */
        XEvent ev;
	char *string;
	char *string2;
	struct stat buf;
	int i, j;
        char str[256];

	/* the default behaviour */

	aspect = 1.0;
        projection = SPHERICAL;
        city_filter = strdup(city_filter0);
        loc_filter = strdup(loc_filter0);

	/* the main struct for the data file */
	scene = (struct imageLayout *) malloc(sizeof(struct imageLayout));
        
	scene->width = 0;
	scene->height = 0;
        scene->im = 0;

	bzero(&scene->o, sizeof(struct options));
	scene->o.zoom = 1.0;
	scene->o.spacing_lat = 15.0;
	scene->o.spacing_lon = 15.0;
	scene->o.desired_categories = 0xFFFFFFFF; /* default is all */
	scene->o.desired_continents = 0xFFFFFFFF; /* default is all */
	scene->o.mapfile = NULL;
	scene->o.outfile = NULL;
	scene->o.datafile = NULL;
	for (i=0; i<=5; i++) {
            scene->o.font[i] = strdup("6x13");
	    xfont[i] = NULL;
	}

	/* initial setting (which is to print nothing !!) */
	for (i=0; i< 5; i++)
	  for (j=0; j<16; j++) {
	      scene->item[i][j].name = NULL;
              scene->item[i][j].bool = INACTIVE;
	  }

	/* parse system wide rcfile */
	scene->o.rcfile = RCFILE;
        parse_rcfile(scene);

	/* Check whether user as specified a non standard rcfile */
	scene->o.rcfile = NULL;
	for (i=1; i<argc-1; i++) 
	  if (!strpcmp(argv[i], "rcfile", 3)) {
	     scene->o.rcfile = strdup(argv[i+1]);
	     break;
	  }
	if (scene->o.rcfile) {
	   if (!stat(scene->o.rcfile, &buf)) 
              parse_rcfile(scene);
	   else
	      fprintf(stderr, "Config rcfile %s does not exist !!\n",
		              scene->o.rcfile);
	} else {   
	   /* if ~/.xrmaprc exists, use that file instead (for a rcfile) */
	   if (scene->o.rcfile) free(scene->o.rcfile);
	   scene->o.rcfile = NULL;
	   string = getenv("HOME");
	   if (string != NULL) {
	      string2 = (char*) malloc(strlen(string) + 12);
	      if (string2 == NULL) {
		 fprintf(stderr,"Malloc error?");
	 	 exit(-3);
	      }
	      strcpy(string2, string);
	      strcat(string2, "/.xrmaprc");
	      if (!stat(string2, &buf)) scene->o.rcfile = string2;
	   }
	   if (scene->o.rcfile) parse_rcfile(scene);
	}

	for (i=0; i<5; i++) {
	    int n;
	    n = 0;
 	    for (j=0; j<16; j++)
	       if (scene->item[i][j].bool != INACTIVE) ++n;
            feature_descr[i] = (char **) malloc((n+1)*sizeof(char **));
	    feature_descr[i][0] = (i==0)? "Configuration" : categ_descr[i];
	    n = 0;
	    for (j=0; j<16; j++)
               if (scene->item[i][j].bool != INACTIVE) {
		  ++n;
                  feature_descr[i][n] = scene->item[i][j].comment;
                  feature_index[i][n] = j;
	       }
	    feature_length[i] = n;
	}

	/* parse the command line options */
	cmd_string = (char*) malloc(1);
	*cmd_string= '\0';
	for (i=1; i<argc; i++) {
	   {
	      j = strlen(cmd_string);
	      cmd_string = (char *) realloc(cmd_string, j+strlen(argv[i])+2);
	      cmd_string[j] = ' ';
	      cmd_string[j+1] = '\0';
	      strcat(cmd_string, argv[i]);
	   }
	}
	if (*cmd_string && parse_cmd_line(scene, cmd_string)) exit(-1);

	cmd_string = (char *) realloc(cmd_string, 1);
	*cmd_string = '\0';
	if (projection==SPHERICAL) aspect = 1.0;

        /* set the window dimensions */
	if (scene->width==0 && scene->height==0) {
	   if (projection==RECTANGULAR) {
	      scene->width = 1000;
	      scene->height = 500 * aspect;
	   } else {
	      scene->width = 640;
	      scene->height = 640 * aspect;
	   }
	} else
	if (scene->height==0) {
	   if (projection==RECTANGULAR) {
	      scene->height = scene->width * (0.5*aspect);
	   } else {
	      scene->height = scene->width * aspect;
	   }
	} else
	if (scene->width==0) {
	   if (projection==RECTANGULAR) {
	      scene->width = scene->height / (0.5*aspect);
	   } else {
	      scene->width = scene->height / aspect;
	   }
	}

	/* set some default values */
        set_radius(scene);
        check_spacing(scene);

	if (scene->o.mapfile == NULL)
	   scene->o.mapfile = (char*) strdup(MAPFILE); /* default */
	if (scene->o.datafile == NULL)
	   scene->o.datafile = (char*) strdup(CITYFILE); /* default */

	parse_datafile(scene);

        /* We are now entering the GUI program */
        runlevel = 1;

	/* Open X display */
        dpy = XOpenDisplay(NULL);
        scr = DefaultScreen(dpy);
        color_depth = DefaultDepth(dpy, scr);
        bigendian = (ImageByteOrder(dpy) == MSBFirst);

        if (color_depth > 16)
            color_pad = 32;
        else if (color_depth > 8)
            color_pad = 16;
        else
            color_pad = 8;

        Root = RootWindow(dpy, scr);
        visual = DefaultVisual(dpy, scr);
	if (color_depth > 8)
           cmap = DefaultColormap(dpy, scr);

        /* Load X fonts */
	for (i=0; i<=5; i++) {
    	   xfont[i] = XLoadQueryFont(dpy, scene->o.font[i]);
	   if (xfont[i] == (XFontStruct *)NULL) {
	      fprintf(stderr, "Can't open font `%s', giving up !!\n", 
                               scene->o.font[i]);
	   exit(1);
	   }
           set_font_params(i);
	}

	/* get all colors from X colormap */
        get_all_colors(scene);

        /* open the X window */
        win = makeWindow(scene);
	if (!(vectordata || img_output) || preview>0.0) XMapWindow(dpy, win);

       	createGCs(scene);

	/* generate the image */
	if (do_menu && !vectordata && !img_output) {
           cmd_mode = 3;
	   cmd_type = 2;
	   image_init(scene);
	   drawWinCmd(scene);
	} else
	   generate(scene);

        /* quit after preview time if user only wants file output */
	if (vectordata || img_output) {
	   if (preview>0.0) {
	      putPicture(scene);
	      usleep((int)(preview*1000000));
	   }
	   exit(0);
        }

	/* check X events */
	for (;;) {
	  XNextEvent(dpy, &ev);
          if (ev.xclient.message_type == wm_protocols &&
              ev.xclient.format == 32 &&
              ev.xclient.data.l[0] == wm_delete_window) exit(0);
	     /* User has just released a key */
	  switch(ev.type) {
	     
	     case KeyRelease : {
                char buffer[1];
                KeySym keysym = 0;
                XLookupString((XKeyEvent *) &ev, buffer, 1, &keysym, NULL);
		/* printf("%d \n", keysym); */
		if (keysym>=XK_KP_0 && keysym<=XK_KP_9)
		   keysym += XK_0-XK_KP_0;
		if (keysym==XK_KP_Add) keysym = XK_plus;
		if (keysym==XK_KP_Subtract) keysym = XK_minus;
		if (keysym==XK_KP_Multiply) keysym = XK_asterisk;
		if (keysym==XK_KP_Divide) keysym = XK_slash;
		if (keysym==XK_KP_Enter) keysym = XK_Return;
		if (keysym==XK_KP_Delete) keysym = XK_BackSpace;
		if (keysym==XK_KP_Decimal) keysym = XK_period;
		if (keysym==XK_BackSpace) keysym = XK_Delete;
		if (cmd_mode==3 || (cmd_mode==4 && cmd_type==5)) {
	           if (keysym!=XK_Escape) break;
                }
		if (cmd_mode) {
		   if (keysym==XK_Escape) {
		      do_escape(scene);
		      break;
		   }
		   else
		   if (keysym==XK_Return) {
		      if (cmd_mode==1 && !parse_cmd_line(scene, cmd_string)) {
			 cmd_mode = 0;
                         show_title(win);
		         generate(scene);
		      }
		      if (cmd_mode==2) {
			 if (cmd_type==1) {
                            #ifdef ZLIB
                            gzFile * fo = NULL;
                            #else
                            FILE * fo = NULL;
                            #endif
                            if (scene->o.outfile) free(scene->o.outfile);
			    scene->o.outfile = strdup(cmd_string);
			    if (overwrite==1) overwrite=-1;
			    openOutfile(scene, &fo);
			    drawWinCmd(scene);
			    if (overwrite==-1) {
			       i = strlen(scene->o.outfile)-4;
			       img_output = 0;
			       if (!strcmp(scene->o.outfile+i, ".ppm"))
				  img_output = 1;
			       if (!strcmp(scene->o.outfile+i, ".xpm"))
				  img_output = 2;
			       if (!strcmp(scene->o.outfile+i+1, ".ps"))
				  img_output = 3;
			       /* otherwise set default format = XPM */
			       if (img_output == 0) img_output = 2;
			       if (img_output <= 2)
			          printImage(fo, scene);
                               closeOutfile(fo);
			       cmd_mode = 0;
			       img_output = 0;
			       show_title(win);
			    }
			 }
			 if (cmd_type==2) {
			    if (scene->o.rcfile) free(scene->o.rcfile);
			    scene->o.rcfile = strdup(cmd_string);
			    parse_rcfile(scene);
			 }
			 if (cmd_type==3) {
			    if (scene->o.datafile) free(scene->o.datafile);
			    scene->o.datafile = strdup(cmd_string);
			    parse_datafile(scene);
			 }
			 if (cmd_type==4) {
			    if (scene->o.mapfile) free(scene->o.mapfile);
			    scene->o.mapfile = strdup(cmd_string);
			 }
			 if (cmd_type>=2) {
			    cmd_mode = 0;
			    cmd_type = 0;
			    show_title(win);
		            generate(scene);
			 }
		      }
		      if ((cmd_mode==4 && cmd_type<=4) || cmd_mode==5) {
			 if (line_clicked>=1 && line_clicked<=box_numitems)
			    str2value(scene, line_clicked, cmd_string);
		      }
                      redisplay(scene, 0);
		   } else {
		      if (keysym==XK_Right) {
		         if (caret_pos<strlen(cmd_string)) ++caret_pos;
		      } 
                      else		      
		      if (keysym==XK_Left) {
		         if (caret_pos>0) --caret_pos;
		      } 
                      else		      		      
                      if (keysym==XK_Delete) {
		         int l, m;
                         l = strlen(cmd_string);
		         if (caret_pos) {
		            for (m=caret_pos-1; m<l; m++)
			       cmd_string[m] = cmd_string[m+1];
			    --caret_pos;
		         }
		      }
		      else
		      if (keysym>=32 && keysym<128) {
		         int l, m;
		         if (cmd_string) 
                            l = strlen(cmd_string);
		         else
			    l = 0;
                         cmd_string = (char *)realloc(cmd_string, l+2);
		         for (m=l; m>caret_pos; m--)
			     cmd_string[m] = cmd_string[m-1];
                         cmd_string[caret_pos] = keysym;
		         cmd_string[l+1] = '\0';
                         ++caret_pos;
		      }
		      drawWinString(scene,cmd_string, BOTTOM);
		   }
		} else
                if (keysym==XK_q || keysym==XK_Q) 
                   exit(0);
		else
		if (keysym==XK_exclam)
		   commands(scene);
                else
		/* was used only for debugging purposes
                if (keysym==XK_less) {
		   parse_datafile(scene);
                   generate(scene);
		}
		else
		*/
		if (keysym==XK_greater)
		   filemenu(scene);
		else
		if (keysym==XK_m || keysym==XK_M) {
		   cmd_mode = 3;
		   show_title(win);
		   drawWinCmd(scene);
		}
                else
                if (keysym==XK_c || keysym==XK_C) {
		   if (scene->item[CONF][C_CITY_CIRCLE].bool==ON &&
                       scene->item[CONF][C_CITY_TEXT].bool==ON) {
                       scene->item[CONF][C_CITY_CIRCLE].bool = OFF; 
                       scene->item[CONF][C_CITY_TEXT].bool = OFF; 
	               drawWinString(scene, 
                         "Generating map without cities.",
                         BOTTOM);
		   } else {
		       if (scene->item[CONF][C_CITY_CIRCLE].bool==OFF) {
                          scene->item[CONF][C_CITY_CIRCLE].bool = ON; 
 	                  drawWinString(scene, 
			    "Generating map with city spots shown.", 
                            BOTTOM);
		       } else {
                          scene->item[CONF][C_CITY_TEXT].bool = ON; 
 	                  drawWinString(scene, 
			    "Generating map with city spots and names shown.", 
                            BOTTOM);
		       }
		   }
		   generate(scene);
                }
		else
                if (keysym==XK_l || keysym==XK_L) {
		   if (scene->item[CONF][C_LOC_LAND].bool==ON || 
                       scene->item[CONF][C_LOC_WATER].bool==ON) {
                      scene->item[CONF][C_LOC_LAND].bool = OFF;
                      scene->item[CONF][C_LOC_WATER].bool = OFF;
	              drawWinString(scene, 
                         "Generating map without locations.", 
                         BOTTOM);
		   } else {
                      scene->item[CONF][C_LOC_LAND].bool = ON;
                      scene->item[CONF][C_LOC_WATER].bool = ON;
	              drawWinString(scene, 
                         "Generating map with locations shown.", 
                         BOTTOM);
		   }
		   generate(scene);
                }
		else
                if (keysym==XK_e || keysym==XK_E) {
		   if (projection == SPHERICAL) {
		      if (scene->item[CONF][C_EARTH_CONTOUR].bool==ON) {
                         scene->item[CONF][C_EARTH_CONTOUR].bool = OFF;
	                 drawWinString(scene, 
                           "Generating map with no Earth contour.", 
                           BOTTOM);
		      } else {
                         scene->item[CONF][C_EARTH_CONTOUR].bool = ON;
	                 drawWinString(scene, 
                          "Generating map with Earth contour.", 
                          BOTTOM);
		      }
		      generate(scene);
		   } else
                      putPicture(scene);
		} 
		else
                if (keysym==XK_g || keysym==XK_G) {
		   if (scene->item[CONF][C_LAT_GRID].bool==ON ||
                       scene->item[CONF][C_LON_GRID].bool==ON) {
                       scene->item[CONF][C_LAT_GRID].bool = OFF; 
                       scene->item[CONF][C_LON_GRID].bool = OFF; 
	               drawWinString(scene, 
                         "Generating map without gridlines.",
                         BOTTOM);
		   } else {
                       scene->item[CONF][C_LAT_GRID].bool = ON; 
                       scene->item[CONF][C_LON_GRID].bool = ON; 
 	               drawWinString(scene, 
                         "Generating map with gridlines shown.", 
                         BOTTOM);
		   }
		   generate(scene);
                }
		else
		if (keysym==XK_t || keysym==XK_T) {
		   if (projection==SPHERICAL) {
	              scene->transparent = 1 - scene->transparent;
		      if (scene->transparent)
		         drawWinString(scene, 
                            "Drawing transparent spherical map.", BOTTOM);
		      else
		         drawWinString(scene, 
                            "Drawing opaque spherical map.", BOTTOM);
		      generate(scene);
		   }
                }
		else
		if (keysym==XK_r || keysym==XK_R) {
	           reverse = 1 - reverse;
		   get_all_colors(scene);
		   drawWinString(scene, "Reversing colors.", BOTTOM);
		   generate(scene);
                }
		else
		if (keysym==XK_plus) {
	           drawWinString(scene, "Zooming map by square root of 2.", BOTTOM);
                   scene->o.zoom *= sqrt(2);
		   generate(scene);
		}
                else
		if (keysym==XK_minus) {
	           drawWinString(scene, "Reducing map by square root of 2.", BOTTOM);
                   scene->o.zoom /= sqrt(2);
		   generate(scene);
		}
		else
		if (keysym==XK_Right) {
                   scene->o.yrot += angle_inc(scene->o.zoom);
	           sprintf(str, "Moving to Longitude %.2f.", scene->o.yrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
		if (keysym==XK_Left) {
                   scene->o.yrot -= angle_inc(scene->o.zoom);
	           sprintf(str, "Moving to Longitude %.2f.", scene->o.yrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
		if (keysym==XK_Up) {
                   scene->o.xrot += angle_inc(scene->o.zoom);
	           sprintf(str, "Moving to Latitude %.2f.", scene->o.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
		if (keysym==XK_Down) {
                   scene->o.xrot -= angle_inc(scene->o.zoom);
	           sprintf(str, "Moving to Latitude %.2f.", scene->o.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
                if (keysym==XK_Home) {
                   scene->o.yrot -= 15.0;
	           sprintf(str, "Moving to Latitude %.2f.", scene->o.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
                if (keysym==XK_End) {
                   scene->o.yrot += 15.0;
	           sprintf(str, "Moving to Latitude %.2f.", scene->o.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
                if (keysym==XK_Page_Up) {
                   scene->o.xrot += 15.0;
	           sprintf(str, "Moving to Latitude %.2f.", scene->o.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
                if (keysym==XK_Page_Down) {
                   scene->o.xrot -= 15.0;
	           sprintf(str, "Moving to Latitude %.2f.", scene->o.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
		if (keysym==XK_Return || keysym==XK_Escape || keysym==XK_space)
		   putPicture(scene);
		else
		if (keysym == XK_h || keysym == XK_H)
		      system("xterm -e man xrmap &");
                else {
		   sprintf(str, "Zoom = %.4f  aspect = %.3f  "
                                "lat = %.3f  lon = %.3f  inc = %.3f",
		   scene->o.zoom, aspect,
                   scene->o.xrot, scene->o.yrot, scene->o.zrot);
                   drawWinString(scene, str, TOP);
                   drawWinString(scene, "Command keys: "
		     "? ! > M Q H C L E G T R + - <Arrows> <Space> <Escape>",
                     BOTTOM);
		}
	     }
	     break;

	     /* User probably has changed window size */
	     case ConfigureNotify : 
	        {
	           int x, y;
		   unsigned int w, h;
		   if (getWindowPlacement(win, &x, &y, &w, &h) &&
                       (w!=scene->width || h!=scene->height) ) {
                      XResizeWindow(dpy, win, w, h);
		      create_textpix(scene->width);
	              sprintf(str, "Resizing window to size %dx%d.", w, h);
		      scene->width = w;
		      scene->height = h;
		      drawWinString(scene, str, BOTTOM);
                      set_radius(scene);
		      XSync(dpy, True);
		      if (cmd_mode==0)
		         generate(scene);
		      else {
			 redisplay(scene, 1);
			 break;
		      }
		   }
	        }

	     /* User has moved the mouse */
	     case MotionNotify :
		if (cmd_mode && ev.xbutton.y<=fonth[0]+5) break;
		if (cmd_mode && ev.xbutton.y>=scene->height-(fonth[0]+6)) break;
                if ((but_pressed==1 || but_pressed==3) && cmd_mode<=1)
		   show_point(scene, ev.xbutton.x, ev.xbutton.y);
	        break;

	     case ButtonPress :
	        but_pressed = ev.xbutton.button;
		if (cmd_mode && ev.xbutton.y<=fonth[0]+5) break;
		if (cmd_mode==3 || (cmd_mode==4 && cmd_type==5)) break;
		if (cmd_mode && ev.xbutton.y>=scene->height-fonth[0]-6) {
		   do_caret(scene, ev.xbutton.x);
		   break;
		}
		if (cmd_mode==0 && city_shown) {
		   city_shown = 0;
		   putPicture(scene);  
		}
		if (cmd_mode==0 && ev.xbutton.button==3) {
	           show_point(scene, ev.xbutton.x, ev.xbutton.y);
		   break;
		}
		if (cmd_mode<3 && ev.xbutton.button<2)
	           show_point(scene, ev.xbutton.x, ev.xbutton.y);
		break;

             case ButtonRelease :
	        but_pressed = 0;
		if (cmd_mode==0 && ev.xbutton.button == 2) {
		   cmd_mode = 3;
		   show_title(win);
		   redisplay(scene, 0);
		   break;
		}
		if (cmd_mode && ev.xbutton.y<=fonth[0]+5) {
		   if (ev.xbutton.x>=scene->width-esc_width) {
		      do_escape(scene);
		      break;
		   }
		   if (cmd_mode<=2) {
                      if (point_shown) {
                         redisplay(scene, 0);
		         break;
		      }
		      if (cmd_type==0) {
		         cmd_type = 5;
		         for (i=5; i>=1; i--) {
		            if (ev.xbutton.x<
                                XTextWidth(xfont[0], MENU2, 13*i-1))
			    cmd_type = i;
			 }
			 if (cmd_type==5) exit(0);
		      } else
			 break;
		   }
		}
		if (cmd_mode && ev.xbutton.y>=scene->height-fonth[0]-6) {
   		   if (cmd_mode==3 || (cmd_mode==4 && cmd_type==5)) break;
		   do_caret(scene, ev.xbutton.x);
		   if (point_shown)
		      redisplay(scene, 0);
		   break;
		}
                if (cmd_mode==3 && ev.xbutton.y<=fonth[0]+5) {
		   int i;
		   cmd_mode = 4;
		   cmd_type = 5;
		   line_clicked = -1;
		   for (i=6; i>=1; i--) {
		      if (ev.xbutton.x<XTextWidth(xfont[0], MENU3, 13*i-1))
			 cmd_type = i-2;
		   }
		   if (cmd_type==-1)
		      filemenu(scene);
		   else
		   if (cmd_type==0) {
		      commands(scene);
                   } else
	              redisplay(scene, 0);
		   break;
		}
                if (cmd_mode==4 && cmd_type==5 && ev.xbutton.y<=fonth[0]+5) {
		   int i;
		   cmd_mode = 5;
		   cmd_type = 5;
		   line_clicked = -1;
		   for (i=4; i>=1; i--) {
		      if (ev.xbutton.x<XTextWidth(xfont[0], MENU4, 18*i-1))
			 cmd_type = i;
		   }
		   redisplay(scene, 1);
		   break;
		}
                if ((cmd_mode==4 && cmd_type<=4) || cmd_mode==5) {
		   int j, l, m;
                   l = -1;
		   if (ev.xbutton.x>box_x && ev.xbutton.x<box_x+box_width) {
		      j = ((int)ev.xbutton.y - box_y);
		      m = 0;
		      while (j<0) { j += box_spacing; m += 1; }
		      j = j/box_spacing - m;
		      if (ev.xbutton.y>box_y+j*box_spacing &&
			  ev.xbutton.y<box_y+j*box_spacing+box_height)
			 l = j;
		   }
		   if (l>=1 && l<=box_numitems) {
		      if (line_clicked == l) {
			 int j=0;
			 if (cmd_mode>=4) {
			    if (cmd_mode==4) {
			       if (cmd_type==1) {
				  if (l==1) {
                                     projection = (projection+1)%4;
			             change_made = 1;
				     cmd_string = (char *) realloc(cmd_string, 3);
				     sprintf(cmd_string, "%d", projection);
				  }
                                  if (l==16) {
				     place_name = 1-place_name;
			             change_made = 1;
				     strcpy(cmd_string, (place_name)? "+":"-");
				  }
                                  if (l==17) {
				     scene->transparent = 1-scene->transparent;
			             change_made = 1;
				     strcpy(cmd_string, 
                                            (scene->transparent)? "+":"-");
				  }
			       }
			       if (cmd_type>=2) {
				  change_made = 1;
                                  j = 1<<(l-1);
			       }
			       if (cmd_type==3) {
				  if (scene->o.desired_continents & j) {
				     scene->o.desired_continents &= ~j;
				     *cmd_string = '-';
				  } else {
				     scene->o.desired_continents |= j;
				     *cmd_string = '+';
				  } 
			       }
			       if (cmd_type==4) {
				  if (scene->o.desired_categories & j) {
				     scene->o.desired_categories &= ~j;
				     *cmd_string = '-';
				  } else {
				     scene->o.desired_categories |= j;
				     *cmd_string = '+';
				  } 
			       }
			    }
			    if (cmd_mode==5) {
			      if (cmd_type>=2 || l>=C_LAT_GRID) {
			         change_made = 1;
			         j = feature_index[cmd_type-1][l];
			         scene->item[cmd_type-1][j].bool =
                                   1-scene->item[cmd_type-1][j].bool;
			         *cmd_string = 
                                     (scene->item[cmd_type-1][j].bool)?
				     '+' : '-';
			      }
			    }
			 }
		      } else {
		        line_clicked = l;
			value2str(scene, line_clicked, str, 0);
		        l = strlen(str);
		        cmd_string = (char *)realloc(cmd_string, l+2);
		        strcpy(cmd_string, str);
		        caret_pos = l;
		      }
		   }
		   redisplay(scene, 0);
		   break;
		}
		if (cmd_mode==3 || (cmd_mode==4 && cmd_type==5)) break;
		if (cmd_mode && ev.xbutton.y>=scene->height-fonth[0]-6) {
                   do_caret(scene, ev.xbutton.x);
	           break;
		}
                if (cmd_mode==0 && ev.xbutton.button==3) {
		   double x, y;
                   if (inverse_coord(scene, 
                       ev.xbutton.x, ev.xbutton.y, &x, &y)) {
		      scene->o.xrot = x;
		      scene->o.yrot = y;
		      drawWinString(scene, "Recentering map to "
                                           "specified point", BOTTOM);
		      generate(scene);
		   } else
		      redisplay(scene, 0);
		} else {
		  if (!city_shown)
		     redisplay(scene, 0);
		}
	        break;

	     /* Expose button events */
	     case Expose :
	        redisplay(scene, 1);
		XSync(dpy, True);
                break;
	     }
	}
	return 0;
}	
