/* xrmap - vector based global map generating program
 *
 * Copyright (C) 2001 Jean-Pierre Demailly <demailly@ujf-grenoble.fr>
 * ftp://ftp.ac-grenoble.fr/ge/geosciences
 *
 * derived from "rmap"
 * Copyright (C) 2000 Reza Naima <reza@reza.net>
 * http://www.reza.net/rmap/
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define PACKAGE "xrmap"
#define VERSION "1.4"
#define DATEPKG "October 16, 2001"

#define MAPFILE SHAREDIR"/CIA_map.rez"
#define CITYFILE SHAREDIR"/Locations"
#define RCFILE SHAREDIR"/Xrmaprc"
#define FLAGPATH SHAREDIR"/flags"

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>


/* The Constants, we just have to have the constants */
#define PI 3.141592654
#define PI_OVER_2 1.57079632679
#define CONV 0.0174532925199         /* PI/180 */
#define ONE_OVER_PI 0.31830988618    /* 1/PI */
#define FIVE_OVER_4PI 0.39788735773  /* 5/4PI */
/* #define EARTH_RADIUS 6378140 */   /* meters at equator */

#define HEADER_MAGIC 0x86460346	/* random number that i chose */
#define BIT32 int		/* OS specific */
#define BIT16 short		/* OS specific */

/* Types of projections */
#define SPHERICAL 0
#define RECTANGULAR 1
#define MERCATOR 2
#define MILLER 3

/* Reminder defines */
#define TOP 0
#define BOTTOM 1
#define PRETOP 2
#define PREBOTTOM 3
#define PREPRETOP 4
#define PREPREBOTTOM 5

/* The following are the Contient Codes */
#define AFRICA 1
#define ASIA 2
#define EUROPE 3
#define NORTH_AMERICA 4
#define SOUTH_AMERICA 5

/* And these are the categories */
#define PBY 1
#define RIV 2
#define BDY 3
#define CIL 5

/* itemdata boolean defines */
#define INACTIVE -1
#define OFF 0
#define ON  1

/* itemdata further defines */
#define CONF 0
#define C_BG_TEXT 1
#define C_FG_TEXT 2
#define C_CARET 3
#define C_ESCAPE 4
#define C_BG_MAP 5
#define C_LAT_GRID 6
#define C_LON_GRID 7
#define C_EARTH_CONTOUR 8
#define C_CITY_CIRCLE 9
#define C_CITY_MARKED 10
#define C_CITY_TEXT 11
#define C_LOC_LAND 12
#define C_LOC_WATER 13

/* structs */
typedef struct
{
  double item[9];
}
matrix;

typedef struct
{
  double item[3];
}
vector;

typedef struct
{
  double item[3];
}
spherePt;

/* And some simple defines to make the vector and matrix structs make sense */
#define X 0
#define Y 1
#define Z 2
#define LONGITUDE 0
#define LATITUDE  1
#define RADIUS    2
/* These defines are for the subroutine to determine if the segment is to be
   viewable or not */
#define NO 0
#define YES 1
#define POINT_ONLY 2


/* and the more important structs... */
struct header
{
  BIT32 magic;
  BIT32 segment_index_address;
  BIT32 segment_index_count;
};

struct segment_index
{
  BIT32 maxlat;
  BIT32 minlat;
  BIT32 maxlong;
  BIT32 minlong;
  BIT32 segment_address;
  BIT32 continent;
  BIT32 category;
  BIT32 type;
};

struct segment_header
{
  BIT32 orgx;
  BIT32 orgy;
  BIT32 nstrokes;
};

struct stroke
{
  BIT32 dx;
  BIT32 dy;
};

struct itemdata
{
        unsigned short r;
        unsigned short g;
        unsigned short b;
        short bool;
        Pixel pix;
        char *name;
        char *comment;
};

struct options 
{
  double spacing_lat;
  double spacing_lon;
  double zoom;
  double xrot;
  double yrot;
  double zrot;
  char* mapfile;
  char* outfile;
  char* rcfile;
  char* datafile;
  char* font[6];
  BIT32 desired_categories;
  BIT32 desired_continents; 
};

/* this struct defines the layout of the image (height, width, perspective..)*/
struct imageLayout
{
  int height;			/* y */
  int width;			/* x */
  short transparent;
  XImage * im;
  struct itemdata item[5][16];
  matrix rotation;
  struct options o;
};

/* timezone struct */

struct timezone
{
        char *code;
        char *flags;
        char *name;
        char *tzvalue;
        int  continent;
};

/* city/location struct */

struct location
{
        char   obj;
        char   status;
        char   drawn;
	char*  country;
	char*  name;
        char*  region;
	double lat;
	double lon;
        int    alt;
        int    rank;
        int    worldrank;
        int    population;
        int    tz;
};

